"""
HuggingFace Hub.

Unlike the other schemes, we don't implement a `SchemeClient` subclass here because
`huggingface_hub` handles the caching logic internally in essentially the same way.
"""

from pathlib import Path
from typing import Optional

import requests

from ..common import PathOrStr
from ..file_lock import FileLock
from ..meta import Meta
from ..util import _lock_file_path, _meta_file_path
from ..version import VERSION


def hf_hub_download(
    url: str, model_identifier: str, filename: Optional[str], cache_dir: PathOrStr
) -> Path:
    import huggingface_hub as hf_hub

    revision: Optional[str]
    if "@" in model_identifier:
        repo_id = model_identifier.split("@")[0]
        revision = model_identifier.split("@")[1]
    else:
        repo_id = model_identifier
        revision = None

    cache_path: Path
    if filename is not None:
        hub_url = hf_hub.hf_hub_url(
            repo_id=repo_id, filename=filename, revision=revision
        )
        cache_path = Path(
            hf_hub.cached_download(
                url=hub_url,
                library_name="cached_path",
                library_version=VERSION,
                cache_dir=cache_dir,
            )
        )
        # HF writes it's own meta '.json' file which uses the same format we used to use and still
        # support, but is missing some fields that we like to have.
        # So we overwrite it when it we can.
        with FileLock(_lock_file_path(cache_path), read_only_ok=True):
            meta = Meta.from_path(_meta_file_path(cache_path))
            # The file HF writes will have 'resource' set to the 'http' URL corresponding to the 'hf://' URL,
            # but we want 'resource' to be the original 'hf://' URL.
            if meta.resource != url:
                meta.resource = url
                meta.to_file()
    else:
        cache_path = Path(
            hf_hub.snapshot_download(repo_id, revision=revision, cache_dir=cache_dir)
        )
        # Need to write the meta file for snapshot downloads if it doesn't exist.
        with FileLock(_lock_file_path(cache_path), read_only_ok=True):
            if not _meta_file_path(cache_path).is_file():
                meta = Meta.new(
                    url,
                    cache_path,
                    extraction_dir=True,
                )
                meta.to_file()
    return cache_path


def hf_get_from_cache(url: str, cache_dir: PathOrStr) -> Path:
    from huggingface_hub.utils import (
        EntryNotFoundError,
        RepositoryNotFoundError,
        RevisionNotFoundError,
    )

    # Remove the 'hf://' prefix
    identifier = url[5:]

    if identifier.count("/") > 1:
        filename = "/".join(identifier.split("/")[2:])
        model_identifier = "/".join(identifier.split("/")[:2])
        return hf_hub_download(url, model_identifier, filename, cache_dir)
    elif identifier.count("/") == 1:
        # 'hf://' URLs like 'hf://xxxx/yyyy' are potentially ambiguous,
        # because this could refer to either:
        #  1. the file 'yyyy' in the 'xxxx' repository, or
        #  2. the repo 'yyyy' under the user/org name 'xxxx'.
        # We default to (1), but if we get a 404 error or 401 error then we try (2)
        try:
            model_identifier, filename = identifier.split("/")
            return hf_hub_download(url, model_identifier, filename, cache_dir)
        except (RepositoryNotFoundError, RevisionNotFoundError, EntryNotFoundError):
            return hf_hub_download(url, identifier, None, cache_dir)
        except requests.exceptions.HTTPError as exc:
            if exc.response is not None and exc.response.status_code in {401, 404}:
                return hf_hub_download(url, identifier, None, cache_dir)
            else:
                raise
        except ValueError:
            return hf_hub_download(url, identifier, None, cache_dir)
    else:
        return hf_hub_download(url, identifier, None, cache_dir)



import glob
import logging
import os
from shutil import copyfile
from typing import Any, Dict, Iterable, List, Optional, Set

from snorkel.classification.multitask_classifier import MultitaskClassifier
from snorkel.types import Config

Metrics = Dict[str, float]


class CheckpointerConfig(Config):
    """Manager for checkpointing model.

    Parameters
    ----------
    checkpoint_dir
        The path to a directory where checkpoints will be saved
        The Trainer will set this to the log directory if it is None
    checkpoint_factor
        Check for a best model every this many evaluations. For example, if
        evaluation_freq is 0.5 epochs and checkpoint_factor is 2, then checkpointing
        will be attempted every 1 epochs.
    checkpoint_metric
        The metric to checkpoint on, of the form "task/dataset/split/metric:mode" where
        mode is "min" or "max".
    checkpoint_task_metrics
        Additional metrics to save best models for. Note that the best model according
        to `checkpoint_metric` will be the one that is loaded after training and used
        for early stopping.
    checkpoint_runway
        No checkpointing will occur for the first this many checkpoint_units
    checkpoint_clear
        If True, clear all checkpoints besides the best so far.
    """

    checkpoint_dir: str = "checkpoints"
    checkpoint_factor: int = 1
    checkpoint_metric: str = "model/all/train/loss:min"
    checkpoint_task_metrics: Optional[List[str]] = None
    checkpoint_runway: int = 0
    checkpoint_clear: bool = True


class Checkpointer:
    """Manager for checkpointing model.

    Parameters
    ----------
    counter_unit
        The unit to use when determining when its time to checkpoint (one of
        ["epochs", "batches", "points"]); must match the counter_unit of LogManager
    evaluation_freq
        How frequently the model is being evaluated (this is the maximum frequency that
        checkpointing can occur, which will happen if checkpoint_factor==1)
    kwargs
        Config merged with ``default_config["checkpointer_config"]``
    """

    def __init__(
        self, counter_unit: str, evaluation_freq: float, **kwargs: Any
    ) -> None:
        self.config = CheckpointerConfig(**kwargs)
        self._validate_config()

        # Pull out checkpoint settings
        self.checkpoint_unit = counter_unit
        self.checkpoint_dir = self.config.checkpoint_dir
        self.checkpoint_clear = self.config.checkpoint_clear
        self.checkpoint_runway = self.config.checkpoint_runway
        self.checkpoint_factor = self.config.checkpoint_factor
        self.checkpoint_condition_met = False

        # Collect all metrics to checkpoint
        self.checkpoint_metric = self._make_metric_map([self.config.checkpoint_metric])
        self.checkpoint_task_metrics = self._make_metric_map(
            self.config.checkpoint_task_metrics
        )
        self.checkpoint_task_metrics.update(self.checkpoint_metric)

        # Create checkpoint directory if necessary
        if not os.path.exists(self.checkpoint_dir):
            os.makedirs(self.checkpoint_dir)

        # Set checkpoint frequency
        self.checkpoint_freq = evaluation_freq * self.checkpoint_factor
        if self.checkpoint_freq <= 0:
            raise ValueError(
                f"Invalid checkpoint freq {self.checkpoint_freq}, "
                f"must be greater than 0."
            )

        logging.info(
            f"Save checkpoints at '{self.checkpoint_dir}' every "
            f"{self.checkpoint_freq} {self.checkpoint_unit}."
        )

        if self.checkpoint_runway > 0:
            logging.info(
                f"No checkpoints will be saved before {self.checkpoint_runway} "
                f"{self.checkpoint_unit}."
            )

        self.best_metric_dict: Dict[str, float] = {}

    def checkpoint(
        self, iteration: float, model: MultitaskClassifier, metric_dict: Metrics
    ) -> None:
        """Check if iteration and current metrics necessitate a checkpoint.

        Parameters
        ----------
        iteration
            Current training iteration
        model
            Model to checkpoint
        metric_dict
            Current performance metrics for model
        """
        # Check if the checkpoint_runway condition is met
        if iteration < self.checkpoint_runway:
            return
        elif not self.checkpoint_condition_met and iteration >= self.checkpoint_runway:
            self.checkpoint_condition_met = True
            logging.info(
                "checkpoint_runway condition has been met. Start checkpointing."
            )

        checkpoint_path = f"{self.checkpoint_dir}/checkpoint_{iteration}.pth"
        model.save(checkpoint_path)
        logging.info(
            f"Save checkpoint at {iteration} {self.checkpoint_unit} "
            f"at {checkpoint_path}."
        )

        if not set(self.checkpoint_task_metrics.keys()).isdisjoint(
            set(metric_dict.keys())
        ):
            new_best_metrics = self._is_new_best(metric_dict)
            for metric in new_best_metrics:
                copyfile(
                    checkpoint_path,
                    f"{self.checkpoint_dir}/best_model_"
                    f"{metric.replace('/', '_')}.pth",
                )

                logging.info(
                    f"Save best model of metric {metric} at {self.checkpoint_dir}"
                    f"/best_model_{metric.replace('/', '_')}.pth"
                )

    def _is_new_best(self, metric_dict: Metrics) -> Set[str]:

        best_metric = set()

        for metric in metric_dict:
            if metric not in self.checkpoint_task_metrics:
                continue
            if metric not in self.best_metric_dict:
                self.best_metric_dict[metric] = metric_dict[metric]
                best_metric.add(metric)
            elif (
                self.checkpoint_task_metrics[metric] == "max"
                and metric_dict[metric] > self.best_metric_dict[metric]
            ):
                self.best_metric_dict[metric] = metric_dict[metric]
                best_metric.add(metric)
            elif (
                self.checkpoint_task_metrics[metric] == "min"
                and metric_dict[metric] < self.best_metric_dict[metric]
            ):
                self.best_metric_dict[metric] = metric_dict[metric]
                best_metric.add(metric)

        return best_metric

    def clear(self) -> None:
        """Clear existing checkpoint files, besides the best-yet model."""
        if self.checkpoint_clear:
            logging.info("Clear all checkpoints other than best so far.")
            file_list = glob.glob(f"{self.checkpoint_dir}/checkpoint_*.pth")
            for fname in file_list:
                os.remove(fname)

    def load_best_model(self, model: MultitaskClassifier) -> MultitaskClassifier:
        """Load the best model from the checkpoint."""
        metric = list(self.checkpoint_metric.keys())[0]
        if metric not in self.best_metric_dict:  # pragma: no cover
            logging.info("No best model found, use the original model.")
        else:
            # Load the best model of checkpoint_metric
            best_model_path = (
                f"{self.checkpoint_dir}/best_model_{metric.replace('/', '_')}.pth"
            )
            logging.info(f"Loading the best model from {best_model_path}.")
            model.load(best_model_path)

        return model

    def _validate_config(self) -> None:
        split_checkpoint_metric = self.config.checkpoint_metric.split("/")
        if len(split_checkpoint_metric) != 4:
            raise ValueError(
                "checkpoint_metric must be formatted 'task/dataset/split/metric:mode'."
            )

        if self.config.checkpoint_runway < 0:
            raise ValueError(
                f"Invalid checkpoint_runway {self.config.checkpoint_runway}, "
                f"must be greater than or equal to 0."
            )

    def _make_metric_map(
        self, metric_mode_iter: Optional[Iterable[str]]
    ) -> Dict[str, str]:
        if metric_mode_iter is None:
            return {}

        metric_mode_map = dict()
        for metric_mode in metric_mode_iter:
            try:
                metric, mode = metric_mode.split(":")
            except ValueError:
                raise ValueError(
                    f"Metric must be of the form 'metric_name:mode' where mode is "
                    f"'max' or 'min'. Instead, got {metric_mode}"
                )
            if mode not in ["min", "max"]:
                raise ValueError(
                    f"Unrecognized checkpoint metric mode {mode} for metric {metric}, "
                    f"must be 'min' or 'max'."
                )
            metric_mode_map[metric] = mode

        return metric_mode_map



import logging
import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt
from .base import set_figure
from ekorpkit import eKonf


log = logging.getLogger(__name__)


def confusion_matrix(ax=None, x=None, y=None, data=None, **kwargs):
    rcParams = {} or kwargs.get(eKonf.Keys.rcPARAMS)
    _ax_cfg_ = {} or kwargs.get("_ax_")
    if ax is None:
        ax = plt.gca()

    log.info(f"Confusion matrix: {kwargs}")

    display_labels = kwargs.get("display_labels") or "auto"
    matrix_labels = kwargs.get("matrix_labels")
    include_values = kwargs.get("include_values")
    include_percentages = kwargs.get("include_percentages")
    summary_stats = kwargs.get("summary_stats")

    # CODE TO GENERATE TEXT INSIDE EACH SQUARE
    blanks = ["" for i in range(data.size)]

    if matrix_labels and len(matrix_labels) == data.size:
        matrix_labels = ["{}\n".format(value) for value in matrix_labels]
    else:
        matrix_labels = blanks

    if include_values:
        matrix_values = ["{0:0.0f}\n".format(value) for value in data.flatten()]
    else:
        matrix_values = blanks

    if include_percentages:
        matrix_percentages = [
            "{0:.2%}".format(value) for value in data.flatten() / np.sum(data)
        ]
    else:
        matrix_percentages = blanks

    box_labels = [
        f"{v1}{v2}{v3}".strip()
        for v1, v2, v3 in zip(matrix_labels, matrix_values, matrix_percentages)
    ]
    box_labels = np.asarray(box_labels).reshape(data.shape[0], data.shape[1])

    # CODE TO GENERATE SUMMARY STATISTICS & TEXT FOR SUMMARY STATS
    if summary_stats:
        # Accuracy is sum of diagonal divided by total observations
        accuracy = np.trace(data) / float(np.sum(data))

        # if it is a binary confusion matrix, show some more stats
        if len(data) == 2:
            # Metrics for Binary Confusion Matrices
            precision = data[1, 1] / sum(data[:, 1])
            recall = data[1, 1] / sum(data[1, :])
            f1_score = 2 * precision * recall / (precision + recall)
            stats_text = "\n\nAccuracy={:0.3f}\nPrecision={:0.3f}\nRecall={:0.3f}\nF1 Score={:0.3f}".format(
                accuracy, precision, recall, f1_score
            )
        else:
            stats_text = "\n\nAccuracy={:0.3f}".format(accuracy)
    else:
        stats_text = ""

    sns.heatmap(
        data,
        annot=box_labels,
        fmt="",
        xticklabels=display_labels,
        yticklabels=display_labels,
        ax=ax,
        **rcParams,
    )

    xlabel = _ax_cfg_.get("xlabel")
    if isinstance(xlabel, str):
        _ax_cfg_["xlabel"] = xlabel + stats_text
    else:
        _ax_cfg_["xlabel"] = stats_text
    set_figure(ax, **_ax_cfg_)



from functools import partial
from typing import Callable, Dict, List, Mapping, Optional, Union

import numpy as np
import pandas as pd

from .metrics import METRICS, metric_score


class Scorer:
    """Calculate one or more scores from user-specified and/or user-defined metrics.

    Parameters
    ----------
    metrics
        A list of metric names, all of which are defined in METRICS
    custom_metric_funcs
        An optional dictionary mapping the names of custom metrics to the functions
        that produce them. Each custom metric function should accept golds, preds, and
        probs as input (just like the standard metrics in METRICS) and return either a
        single score (float) or a dictionary of metric names to scores (if the function
        calculates multiple values, for example). See the unit tests for an example.
    abstain_label
        The gold label for which examples will be ignored. By default, follow convention
        that abstains are -1.

    Raises
    ------
    ValueError
        If a specified standard metric is not found in the METRICS dictionary

    Attributes
    ----------
    metrics
        A dictionary mapping metric names to the corresponding functions for calculating
        that metric
    """

    def __init__(
        self,
        metrics: Optional[List[str]] = None,
        custom_metric_funcs: Optional[Mapping[str, Callable[..., float]]] = None,
        abstain_label: Optional[int] = -1,
    ) -> None:

        self.metrics: Dict[str, Callable[..., float]]
        self.metrics = {}
        if metrics:
            for metric in metrics:
                if metric not in METRICS:
                    raise ValueError(f"Unrecognized metric: {metric}")

                filter_dict = (
                    {}
                    if abstain_label is None or metric == "coverage"
                    else {"golds": [abstain_label], "preds": [abstain_label]}
                )
                self.metrics.update(
                    {
                        metric: partial(
                            metric_score, metric=metric, filter_dict=filter_dict
                        )
                    }
                )

        if custom_metric_funcs is not None:
            self.metrics.update(custom_metric_funcs)

        self.abstain_label = abstain_label

    def score(
        self,
        golds: np.ndarray,
        preds: Optional[np.ndarray] = None,
        probs: Optional[np.ndarray] = None,
    ) -> Dict[str, float]:
        """Calculate scores for one or more user-specified metrics.

        Parameters
        ----------
        golds
            An array of gold (int) labels to base scores on
        preds
            An [n_datapoints,] or [n_datapoints, 1] array of (int) predictions to score
        probs
            An [n_datapoints, n_classes] array of probabilistic (float) predictions

        Because most metrics require either `preds` or `probs`, but not both, these
        values are optional; it is up to the metric function that will be called  to
        raise an exception if a field it requires is not passed to the `score()` method.

        Returns
        -------
        Dict[str, float]
            A dictionary mapping metric names to metric scores

        Raises
        ------
        ValueError
            If no gold labels were provided
        """
        if len(golds) == 0:  # type: ignore
            raise ValueError("Cannot score empty labels")

        metric_dict = dict()

        for metric_name, metric in self.metrics.items():
            score = metric(golds, preds, probs)
            if isinstance(score, dict):
                metric_dict.update(score)
            else:
                metric_dict[metric_name] = score

        return metric_dict

    def score_slices(
        self,
        S: np.recarray,
        golds: np.ndarray,
        preds: np.ndarray,
        probs: np.ndarray,
        as_dataframe: bool = False,
    ) -> Union[Dict[str, Dict[str, float]], pd.DataFrame]:
        """Calculate user-specified and/or user-defined metrics overall + slices.

        Parameters
        ----------
        S
            A recarray with entries of length n_examples corresponding to slice names
        golds
            Gold (aka ground truth) labels (integers)
        preds
            Predictions (integers)
        probs:
            Probabilities (floats)
        as_dataframe
            A boolean indicating whether to return results as pandas ``DataFrame`` (True)
            or dict (False)

        Returns
        -------
        Union[Dict, pd.DataFrame]
            A dictionary mapping slice_name to metric names to metric scores
            or metrics formatted as pandas ``DataFrame``
        """

        correct_shapes = S.shape[0] == len(golds) == len(preds) == len(probs)
        if not correct_shapes:
            raise ValueError(
                "S, golds, preds, and probs must have the same number of elements"
            )

        # Include overall metrics
        metrics_dict = {"overall": self.score(golds, preds, probs)}

        # Include slice metrics
        slice_names = S.dtype.names
        for slice_name in slice_names:
            mask = S[slice_name].astype(bool)
            metrics_dict[slice_name] = self.score(golds[mask], preds[mask], probs[mask])

        if as_dataframe:
            return pd.DataFrame.from_dict(metrics_dict).transpose()

        return metrics_dict



from typing import List, Mapping, Optional

import torch
import torch.nn.functional as F

Outputs = Mapping[str, List[torch.Tensor]]


def cross_entropy_with_probs(
    input: torch.Tensor,
    target: torch.Tensor,
    weight: Optional[torch.Tensor] = None,
    reduction: str = "mean",
) -> torch.Tensor:
    """Calculate cross-entropy loss when targets are probabilities (floats), not ints.

    PyTorch's F.cross_entropy() method requires integer labels; it does accept
    probabilistic labels. We can, however, simulate such functionality with a for loop,
    calculating the loss contributed by each class and accumulating the results.
    Libraries such as keras do not require this workaround, as methods like
    "categorical_crossentropy" accept float labels natively.

    Note that the method signature is intentionally very similar to F.cross_entropy()
    so that it can be used as a drop-in replacement when target labels are changed from
    from a 1D tensor of ints to a 2D tensor of probabilities.

    Parameters
    ----------
    input
        A [num_points, num_classes] tensor of logits
    target
        A [num_points, num_classes] tensor of probabilistic target labels
    weight
        An optional [num_classes] array of weights to multiply the loss by per class
    reduction
        One of "none", "mean", "sum", indicating whether to return one loss per data
        point, the mean loss, or the sum of losses

    Returns
    -------
    torch.Tensor
        The calculated loss

    Raises
    ------
    ValueError
        If an invalid reduction keyword is submitted
    """
    num_points, num_classes = input.shape
    # Note that t.new_zeros, t.new_full put tensor on same device as t
    cum_losses = input.new_zeros(num_points)
    for y in range(num_classes):
        target_temp = input.new_full((num_points,), y, dtype=torch.long)
        y_loss = F.cross_entropy(input, target_temp, reduction="none")
        if weight is not None:
            y_loss = y_loss * weight[y]
        cum_losses += target[:, y].float() * y_loss

    if reduction == "none":
        return cum_losses
    elif reduction == "mean":
        return cum_losses.mean()
    elif reduction == "sum":
        return cum_losses.sum()
    else:
        raise ValueError("Keyword 'reduction' must be one of ['none', 'mean', 'sum']")



from typing import Sequence

from snorkel.classification.data import DictDataLoader

from .scheduler import BatchIterator, Scheduler


class SequentialScheduler(Scheduler):
    """Return batches from all dataloaders in sequential order."""

    def __init__(self) -> None:
        super().__init__()

    def get_batches(self, dataloaders: Sequence[DictDataLoader]) -> BatchIterator:
        """Return batches from dataloaders sequentially in the order they were given.

        Parameters
        ----------
        dataloaders
            A sequence of dataloaders to get batches from

        Yields
        ------
        (batch, dataloader)
            batch is a tuple of (X_dict, Y_dict) and dataloader is the dataloader
            that that batch came from. That dataloader will not be accessed by the
            model; it is passed primarily so that the model can pull the necessary
            metadata to know what to do with the batch it has been given.
        """
        for dataloader in dataloaders:
            for batch in dataloader:
                yield batch, dataloader



import logging
from .json import load_json


log = logging.getLogger(__name__)


def mail_to_json(contents):
    try:
        import mailparser
    except ImportError:
        log.warning(
            "mailparser is not installed. Please install it with `pip install mailparser`"
        )
    try:
        contents = mailparser.parse_from_string(contents).mail_json
    except Exception as err:
        print(f"{contents}")
        print(err)
        contents = None
    return load_json(contents)



import requests
import wget
import logging
import os
from bs4 import BeautifulSoup


log = logging.getLogger(__name__)


class ExPORTER:
    def __init__(self, output_dir, force_download=False, **kwargs):
        self.force_download = force_download
        self.output_dir = output_dir

        self.download_urls = {
            "exporter": "https://exporter.nih.gov/ExPORTER_Catalog.aspx?sid=0&index=1",
            "crisp": "https://exporter.nih.gov/CRISP_Catalog.aspx?sid=0&index=1",
        }
        self.targets = {
            "exporter": "CSVs/final/RePORTER_",
            "crisp": "CRISP/Abstracts/CRISP_PRJABS_C_",
        }
        self.base_url = "https://exporter.nih.gov/"
        self.download()

    def download(self):
        for subset in self.targets.keys():
            self.download_files(subset)

    def download_files(self, subset):
        log.info(f"\nsubset: {subset}\n")
        output_dir = self.output_dir + "/" + subset
        if not os.path.exists(output_dir):
            os.makedirs(output_dir, exist_ok=True)
        if os.listdir(output_dir) and not self.force_download:
            log.info("Files already downloaded. Skipping.")
            return

        url = self.download_urls[subset]
        target = self.targets[subset]
        self.sess = requests.session()
        r = self.sess.get(url)
        soup = BeautifulSoup(r.content, "lxml")
        links = [a["href"] for a in soup.find_all("a", href=True)]
        links = [link for link in links if target in link]

        log.info(f"Found {len(links)} links in Exporter")
        for i, link in enumerate(links):
            filename = os.path.basename(link)
            filepath = os.path.join(output_dir, filename)
            if not os.path.isfile(filepath) or os.path.getsize(filepath) == 0:
                log.info(f"\nDownloading file {i+1} of {len(links)}: {link}")
                wget.download(self.base_url + link, filepath)
            else:
                log.info(f"File {filename} already exists. Skipping.")



import logging
from collections import defaultdict
from typing import DefaultDict, Dict, List, Tuple

import numpy as np

from ..utils import to_int_label_array


def get_label_buckets(*y: np.ndarray) -> Dict[Tuple[int, ...], np.ndarray]:
    """Return data point indices bucketed by label combinations.

    Parameters
    ----------
    *y
        A list of np.ndarray of (int) labels

    Returns
    -------
    Dict[Tuple[int, ...], np.ndarray]
        A mapping of each label bucket to a NumPy array of its corresponding indices

    Example
    -------
    A common use case is calling ``buckets = label_buckets(Y_gold, Y_pred)`` where
    ``Y_gold`` is a set of gold (i.e. ground truth) labels and ``Y_pred`` is a
    corresponding set of predicted labels.

    >>> Y_gold = np.array([1, 1, 1, 0])
    >>> Y_pred = np.array([1, 1, -1, -1])
    >>> buckets = get_label_buckets(Y_gold, Y_pred)

    The returned ``buckets[(i, j)]`` is a NumPy array of data point indices with
    true label i and predicted label j.

    More generally, the returned indices within each bucket refer to the order of the
    labels that were passed in as function arguments.

    >>> buckets[(1, 1)]  # true positives
    array([0, 1])
    >>> (1, 0) in buckets  # false positives
    False
    >>> (0, 1) in buckets  # false negatives
    False
    >>> (0, 0) in buckets  # true negatives
    False
    >>> buckets[(1, -1)]  # abstained positives
    array([2])
    >>> buckets[(0, -1)]  # abstained negatives
    array([3])
    """
    buckets: DefaultDict[Tuple[int, int], List[int]] = defaultdict(list)
    y_flat = list(map(lambda x: to_int_label_array(x, flatten_vector=True), y))
    if len(set(map(len, y_flat))) != 1:
        raise ValueError("Arrays must all have the same number of elements")
    for i, labels in enumerate(zip(*y_flat)):
        buckets[labels].append(i)
    return {k: np.array(v) for k, v in buckets.items()}


def get_label_instances(
    bucket: Tuple[int, ...], x: np.ndarray, *y: np.ndarray
) -> np.ndarray:
    """Return instances in x with the specified combination of labels.

    Parameters
    ----------
    bucket
        A tuple of label values corresponding to which instances from x are returned
    x
        NumPy array of data instances to be returned
    *y
        A list of np.ndarray of (int) labels

    Returns
    -------
    np.ndarray
        NumPy array of instances from x with the specified combination of labels

    Example
    -------
    A common use case is calling ``get_label_instances(bucket, x.to_numpy(), Y_gold, Y_pred)``
    where ``x`` is a NumPy array of data instances that the labels correspond to,
    ``Y_gold`` is a list of gold (i.e. ground truth) labels, and
    ``Y_pred`` is a corresponding list of predicted labels.

    >>> import pandas as pd
    >>> x = pd.DataFrame(data={'col1': ["this is a string", "a second string", "a third string"], 'col2': ["1", "2", "3"]})
    >>> Y_gold = np.array([1, 1, 1])
    >>> Y_pred = np.array([1, 0, 0])
    >>> bucket = (1, 0)

    The returned NumPy array of data instances from ``x`` will correspond to
    the rows where the first list had a 1 and the second list had a 0.
    >>> get_label_instances(bucket, x.to_numpy(), Y_gold, Y_pred)
    array([['a second string', '2'],
           ['a third string', '3']], dtype=object)

    More generally, given bucket ``(i, j, ...)`` and lists ``y1, y2, ...``
    the returned data instances from ``x`` will correspond to the rows where
    y1 had label i, y2 had label j, and so on. Note that ``x`` and ``y``
    must all be the same length.
    """
    if len(y) != len(bucket):
        raise ValueError("Number of lists must match the amount of labels in bucket")
    if x.shape[0] != len(y[0]):
        # Note: the check for all y having the same number of elements occurs in get_label_buckets
        raise ValueError(
            "Number of rows in x does not match number of elements in at least one label list"
        )
    buckets = get_label_buckets(*y)
    try:
        indices = buckets[bucket]
    except KeyError:
        logging.warning("Bucket" + str(bucket) + " does not exist.")
        return np.array([])
    instances = x[indices]
    return instances



import os
import pandas as pd
import omegaconf
import logging
import platform
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from matplotlib import font_manager, rc
from pathlib import Path
from ekorpkit import eKonf


log = logging.getLogger(__name__)


def prepare_data(data, _query=None, _set_index=None, **kwargs):
    data = data.copy()
    if _query is not None:
        if isinstance(data, pd.DataFrame):
            data = data.query(_query, engine="python")
        else:
            data = [d for d in data if eval(_query)]
    if _set_index is not None:
        if isinstance(data, pd.DataFrame):
            if _set_index == "reset":
                data = data.reset_index()
            else:
                data = data.set_index(_set_index)
    return data


def prepare_plot_args(data, verbose=False, **kwargs):
    kwargs = eKonf.to_dict(kwargs)
    _plots = kwargs.get("plots")
    _axes = kwargs.get("axes")
    _figure = kwargs.get("figure") or {}
    _style = kwargs.get("style") or {}
    _gridspec = kwargs.get("gridspec") or {}
    _subplots = kwargs.get("subplots") or {}
    _savefig = kwargs.get("savefig") or {}
    _theme = kwargs.get("theme") or {}
    _grid = kwargs.get("grid") or {}
    _fig_super = _figure.pop("super", None) or {}
    _fig_rcParams = _figure.pop("rcParams", None) or {}

    _style["rcParams"] = _style.get("rcParams") or {}
    _style["rcParams"].update(_fig_rcParams)
    _gridspec["nrows"] = _subplots.pop("nrows", 1)
    _gridspec["ncols"] = _subplots.pop("ncols", 1)

    if data is None:
        log.info("No data to plot")
    if isinstance(data, omegaconf.listconfig.ListConfig):
        data = list(data)
    if verbose:
        log.info(f"type of data: {type(data)}")

    tight_layout = _figure.pop("tight_layout", True)
    figsize = _figure.pop("figsize", None)
    if figsize is not None and isinstance(figsize, str):
        figsize = eval(figsize)

    return (
        _axes,
        _fig_super,
        _grid,
        _gridspec,
        _plots,
        _savefig,
        _style,
        _subplots,
        _theme,
        data,
        figsize,
        tight_layout,
    )


def save_figure(fig, fname=None, **kwargs):
    if fname:
        Path(fname).parent.mkdir(parents=True, exist_ok=True)
        fig.savefig(fname=fname, **kwargs)
        log.info(f"Saved figure to {fname}")


def add_decorations(ax=None, **kwargs):
    if ax is None:
        ax = plt.gca()
    axvspans = [] or kwargs.get("axvspans")
    annotations = [] or kwargs.get("annotations")

    if isinstance(axvspans, dict):
        axvspans = [axvspans]
    if isinstance(annotations, dict):
        annotations = [annotations]

    for span in axvspans:
        if isinstance(span, dict):
            ax.axvspan(**span)
    for annot in annotations:
        if isinstance(annot, dict):
            x = annot.pop("x")
            y = annot.pop("y")
            if x and y:
                annot["xy"] = (x, y)
            xtext = annot.pop("xtext")
            ytext = annot.pop("ytext")
            if xtext and ytext:
                annot["xytext"] = (xtext, ytext)
            ax.annotate(**annot)


def set_super(
    fig,
    xlabel=None,
    ylabel=None,
    title=None,
    **kwargs,
):
    if xlabel is not None:
        if isinstance(xlabel, str):
            fig.supxlabel(xlabel)
        else:
            fig.supxlabel(**xlabel)
    if ylabel is not None:
        if isinstance(ylabel, str):
            fig.supylabel(ylabel)
        else:
            fig.supylabel(**ylabel)
    if title is not None:
        if isinstance(title, str):
            fig.suptitle(title)
        else:
            fig.suptitle(**title)


def set_figure(
    ax,
    xlabel=None,
    ylabel=None,
    title=None,
    legend=None,
    grid=None,
    xlim=None,
    ylim=None,
    xticks=None,
    yticks=None,
    xticklabels=None,
    yticklabels=None,
    xtickmajorformatterfunc=None,
    ytickmajorformatterfunc=None,
    xtick_params=None,
    ytick_params=None,
    **kwargs,
):
    if xlabel is not None:
        if isinstance(xlabel, str):
            ax.set_xlabel(xlabel)
        else:
            ax.set_xlabel(**xlabel)
    if ylabel is not None:
        if isinstance(ylabel, str):
            ax.set_ylabel(ylabel)
        else:
            ax.set_ylabel(**ylabel)
    if title is not None:
        if isinstance(title, str):
            ax.set_title(title)
        else:
            ax.set_title(**title)
    if legend is not None:
        if isinstance(legend, (str, list)):
            ax.legend(legend)
        elif legend.get("labels") is not None:
            ax.legend(**legend)
    if grid is not None:
        if isinstance(grid, bool):
            ax.grid(grid)
        elif isinstance(grid, dict):
            ax.grid(**grid)
    if xticks is not None:
        if isinstance(xticks, str):
            ax.set_xticks(eval(xticks))
        else:
            if xticks.get("labels", None) or xticks.get("ticks", None):
                ax.set_xticks(**xticks)
    if yticks is not None:
        if isinstance(yticks, str):
            ax.set_yticks(eval(yticks))
        else:
            if yticks.get("labels", None) or yticks.get("ticks", None):
                ax.set_yticks(**yticks)
    if xticklabels is not None:
        if not xticklabels.get("labels", None):
            xticklabels["labels"] = ax.get_xticks().tolist()
        ax.set_xticklabels(**xticklabels)
    if yticklabels is not None:
        if not yticklabels.get("labels", None):
            yticklabels["labels"] = ax.get_yticks().tolist()
        ax.set_yticklabels(**yticklabels)
    if xlim is not None:
        if isinstance(xlim, str):
            ax.set_xlim(eval(xlim))
        else:
            ax.set_xlim(**xlim)
    if ylim is not None:
        if isinstance(ylim, str):
            ax.set_ylim(eval(ylim))
        else:
            ax.set_ylim(**ylim)
    if xtickmajorformatterfunc is not None:
        ax.xaxis.set_major_formatter(eval(xtickmajorformatterfunc))
    if ytickmajorformatterfunc is not None:
        ax.yaxis.set_major_formatter(eval(ytickmajorformatterfunc))
    if xtick_params is not None:
        ax.tick_params(axis="x", **xtick_params)
    if ytick_params is not None:
        ax.tick_params(axis="y", **ytick_params)


def set_style(style, rcParams, fontpath=None, language=None, **kwargs):
    if language or (fontpath and Path(fontpath).is_file()):
        fontname, fontpath = _configure_font(fontpath=fontpath)
        if fontname:
            rcParams["font.family"] = fontname

    plt.style.use(style)
    plt.rcParams.update(rcParams)


def find_font_file(query):
    matches = list(
        filter(
            lambda path: query in os.path.basename(path), font_manager.findSystemFonts()
        )
    )
    return matches


def _configure_font(
    set_font_for_matplot=True, fontpath=None, fontname=None, verbose=False
):
    if fontname and not fontname.endswith(".ttf"):
        fontname += ".ttf"
    if not fontpath:
        if platform.system() == "Darwin":
            fontname = fontname or "AppleGothic.ttf"
            fontpath = os.path.join("/System/Library/Fonts/Supplemental/", fontname)
        elif platform.system() == "Windows":
            fontname = fontname or "malgun.ttf"
            fontpath = os.path.join("c:/Windows/Fonts/", fontname)
        elif platform.system() == "Linux":
            fontname = fontname or "NanumGothic.ttf"
            if fontname.lower().startswith("nanum"):
                fontpath = os.path.join("/usr/share/fonts/truetype/nanum/", fontname)
            else:
                fontpath = os.path.join("/usr/share/fonts/truetype/", fontname)
        if fontpath and not Path(fontpath).is_file():
            paths = find_font_file(fontname)
            if len(paths) > 0:
                fontpath = paths[0]
            else:
                fontpath = None
        if verbose:
            log.info(f"Font path: {fontpath}")

    if fontpath is None or not Path(fontpath).is_file():
        fontname = None
        fontpath = None
        log.warning(f"Font file does not exist at {fontpath}")
        if platform.system() == "Linux":
            font_install_help = """
            apt install fontconfig
            apt install fonts-nanum
            fc-list | grep -i nanum
            """
            print(font_install_help)
    else:
        font_manager.fontManager.addfont(fontpath)
        fontname = font_manager.FontProperties(fname=fontpath).get_name()

        if set_font_for_matplot and fontname:
            rc("font", family=fontname)
            plt.rcParams["axes.unicode_minus"] = False
            font_family = plt.rcParams["font.family"]
            if verbose:
                log.info(f"font family: {font_family}")
        if verbose:
            log.info(f"font name: {fontname}")
    return fontname, fontpath



import logging
from typing import Any, Optional

from snorkel.classification.multitask_classifier import MultitaskClassifier
from snorkel.types import Config

from .checkpointer import Checkpointer
from .log_writer import LogWriter


class LogManagerConfig(Config):
    """Manager for checkpointing model.

    Parameters
    ----------
    counter_unit
        The unit to use when assessing when it's time to log.
        Options are ["epochs", "batches", "points"]
    evaluation_freq:
        Evaluate performance on the validation set every this many counter_units
    """

    counter_unit: str = "epochs"
    evaluation_freq: float = 1.0


class LogManager:
    """A class to manage logging during training progress.

    Parameters
    ----------
    n_batches_per_epoch
        Total number batches per epoch
    log_writer
        ``LogWriter`` for current run logs
    checkpointer
        ``Checkpointer`` for current model
    kwargs
        Settings to update in LogManagerConfig
    """

    def __init__(
        self,
        n_batches_per_epoch: int,
        log_writer: Optional[LogWriter] = None,
        checkpointer: Optional[Checkpointer] = None,
        **kwargs: Any,
    ) -> None:
        self.config = LogManagerConfig(**kwargs)
        self.n_batches_per_epoch = n_batches_per_epoch

        self.log_writer = log_writer
        self.checkpointer = checkpointer

        # Set up counter unit
        self.counter_unit = self.config.counter_unit
        if self.counter_unit not in ["points", "batches", "epochs"]:
            raise ValueError(f"Unrecognized counter_unit: {self.counter_unit}")

        # Set up evaluation frequency
        self.evaluation_freq = self.config.evaluation_freq
        logging.info(f"Evaluating every {self.evaluation_freq} {self.counter_unit}.")

        # Set up number of X passed since last evaluation/checkpointing and total
        self.point_count = 0
        self.point_total = 0

        self.batch_count = 0
        self.batch_total = 0

        self.epoch_count = 0.0
        self.epoch_total = 0.0

        self.unit_count = 0.0
        self.unit_total = 0.0

        # Set up count that triggers the evaluation since last checkpointing
        self.trigger_count = 0

    def update(self, batch_size: int) -> None:
        """Update the count and total number."""

        # Update number of points
        self.point_count += batch_size
        self.point_total += batch_size

        # Update number of batches
        self.batch_count += 1
        self.batch_total += 1

        # Update number of epochs
        self.epoch_count = self.batch_count / self.n_batches_per_epoch
        self.epoch_total = self.batch_total / self.n_batches_per_epoch

        # Update number of units
        if self.counter_unit == "points":
            self.unit_count = self.point_count
            self.unit_total = self.point_total
        if self.counter_unit == "batches":
            self.unit_count = self.batch_count
            self.unit_total = self.batch_total
        elif self.counter_unit == "epochs":
            self.unit_count = self.epoch_count
            self.unit_total = self.epoch_total

    def trigger_evaluation(self) -> bool:
        """Check if current counts trigger evaluation."""
        satisfied = self.unit_count >= self.evaluation_freq
        if satisfied:
            self.trigger_count += 1
            self.reset()
        return satisfied

    def trigger_checkpointing(self) -> bool:
        """Check if current counts trigger  checkpointing."""
        if self.checkpointer is None:
            return False
        satisfied = self.trigger_count >= self.checkpointer.checkpoint_factor
        if satisfied:
            self.trigger_count = 0
        return satisfied

    def reset(self) -> None:
        """Reset counters."""
        self.point_count = 0
        self.batch_count = 0
        self.epoch_count = 0
        self.unit_count = 0

    def cleanup(self, model: MultitaskClassifier) -> MultitaskClassifier:
        """Close the log writer and checkpointer if needed. Reload best model."""
        if self.log_writer is not None:
            self.log_writer.cleanup()
        if self.checkpointer is not None:
            self.checkpointer.clear()
            model = self.checkpointer.load_best_model(model)
        return model



import os
import pandas as pd

# from pandarallel import pandarallel
from datetime import datetime, timedelta
import re

from ekorpkit.io.db import connect_mongo
from ekorpkit.preprocessors.normalizer import repeat_normalize


class SecReport:
    REMOVE_PATTERNS_1 = [
        # 한글, 영어, 띄어쓰기, 일부 특수 문자 등을 제외하고 모두 제거
        r"[^ .,?!/@$%~％·∼()\x00-\x7F가-힣]+",
        # 아래 두개는 각주 형식이라서 있으면 이상하게 문자열 끝에 추가됨;;
        r"\[\*[^\]]+\]",
        r"~~[^~]+~~",
    ]
    REMOVE_PATTERNS_2 = [
        # EMAIL_PATTERN
        r"[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+",
        # URL_PATTERN
        r"(ftp|http|https)?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\(\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+",
        # HTML tags
        r"<[a-zA-Z]*?>",
        # 빈 괄호 제거
        r"(\(|\[|<)+[^가-힣A-Za-z\d]*(>|\]|\))+",
        # (, xxxxx) 에서 , 제거
        r"(?<=[(\(|\[|<)+])\s*(\s*,\s*)+\s*",
        # [*** Not working ***] (xxxxx , ) 에서 , 제거
        # r'\s*(\s*,\s*)+\s*(?=[(>|\]|\))+]+)'
        # 괄호와 그 내용들 제거, 안녕(하세요) -> 안녕
        # r"\([^\)]*\)"
    ]

    REPLACE_PATTERNS = {
        # \n -> 띄어쓰기
        # '\\n': "\n",
        # \' -> '
        # "\\'": "'",
        # 빈 괄호 제거
        r"(\(|\[|<)+\s*(>|\]|\))+": "",
        # MULTIPLE_SPACES
        r" +": " ",
        # WIKI_SPACE_CHARS
        r"(\\s|゙|゚|　)+": " ",
        # MULTIPLE_NEWLINES
        r"(?<!\s)(?<!\.)\s*\n\s+\n": "\n",
        r"(?<!\w\.\w.)(?<![A-Z][a-z]\.)(?<=[다]\.|[다까]\?|[다]\!|[가-힣A-Za-z\s]=)\s+": "\n"
        # r'\s*\n\s*\n\s*': '  \t'
    }

    SKIP_SENTENCE_PATTERNS = {
        # WIKI_REMOVE_CHARS
        r"'+|(=+.{2,30}=+)|__TOC__|(ファイル:).+|:(en|de|it|fr|es|kr|zh|no|fi):",
        # [글 = 정진건 기자], ???  < 논설위원 겸 경제교육연구소장 >
        # r'(\[|\(|<|)\s?(글\s?=\s?)?([가-힣A-Za-z·, ]{2,6}\s*){1,2}(기자|특파원),?\s*(\]|\)|>|.?=|)(?![가-힣])'
        r"\[[^\]]*(저작권자|ⓒ|=|기자|특파원)[^\]]*\]",
        r"\([^\)]*(저작권자|ⓒ|=|기자|특파원)[^\)]*\)",
        r"<[^>]*(저작권자|ⓒ|=|기자|특파원)[^>]*>",
        r"(본|이).?기사는.+기사.*입니다",
        r"무단전재.*금지",
        # Wiki section
        r"Section:+",
    }

    def __init__(
        self,
        name,
        output_dir,
        output_file,
        limit=0,
        mongo=None,
        debug_mode=False,
        num_workers=1,
        normalize=None,
        force_download=False,
        **kwargs,
    ):
        self.name = name
        self.reports_df = None
        self.output_path = f"{output_dir}/{output_file}"
        os.makedirs(output_dir, exist_ok=True)
        self.limit = limit
        self.num_workers = num_workers
        self.debug_mode = debug_mode
        self.min_len = normalize.min_len
        self.min_words = normalize.min_words
        self.num_repeats = normalize.num_repeats

        if os.path.exists(self.output_path) and not force_download:
            print(f"output file {self.output_path} already exists, skipping")
        else:
            self.mongo = connect_mongo(**mongo)
            self.results = self._iter_reports()
            self.extract_reports()

    def extract_reports(self):
        reports_text = []
        for rst in self.results:
            # print(rst)
            doc = self._preprocess_report(rst)
            # print(doc)]
            if doc:
                reports_text.append(doc)
        # Store the output in compressed format
        self.reports_df = pd.DataFrame(reports_text)
        self.normalize()
        self.reports_df.to_csv(self.output_path, header=True)
        print(self.reports_df.tail())
        print(f"Corpus [{self.name}] is built to [{self.output_path}]")

    def normalize(self):
        text_key = "text"

        def normalize(row):
            text = row[text_key]
            return self._normalize_text(text)

        # if self.num_workers > 1:
        # 	pandarallel.initialize(nb_workers=self.num_workers)
        # elif self.num_workers < 1:
        # 	pandarallel.initialize()

        df = self.reports_df
        # if self.num_workers == 1:
        df[text_key] = df.apply(normalize, axis=1)
        # else:
        # 	df[text_key] = df.parallel_apply(normalize, axis=1)

        self.reports_df = df.dropna()
        print("Normalization complete!")

    def _normalize_text(self, text):
        # remove non-UTF
        text = text.encode("utf-8", "ignore").decode()

        for pattern in self.REMOVE_PATTERNS_1:
            regex = re.compile(pattern, re.IGNORECASE | re.MULTILINE)
            text = re.sub(regex, "", text)
        for pattern in self.REMOVE_PATTERNS_2:
            regex = re.compile(pattern, re.IGNORECASE | re.MULTILINE)
            text = re.sub(regex, "", text)
        for pattern, v in self.REPLACE_PATTERNS.items():
            regex = re.compile(pattern, re.IGNORECASE | re.MULTILINE)
            text = re.sub(regex, v, text)

        paragraphs = []
        for paragraph in re.split(r"\n\n", text):
            sentences = []
            for s in re.split(r"\n", paragraph):
                skip = False
                for pattern in self.SKIP_SENTENCE_PATTERNS:
                    regex = re.compile(pattern, re.IGNORECASE)
                    pos = re.search(regex, s)
                    if pos:
                        skip = True
                        break

                if (
                    len(s) >= self.min_len
                    and len(s.split()) >= self.min_words
                    and not skip
                ):
                    if self.num_repeats > 1:
                        s = repeat_normalize(s, num_repeats=self.num_repeats)
                    sentences.append(s.strip())
            if len(sentences) > 0:
                paragraphs.append(sentences)
        if len(paragraphs) > 0:
            text = "\n\n".join(["\n".join(p) for p in paragraphs])
        else:
            text = None
        return text

    def _iter_reports(self):
        for item in self.mongo.find({"DT": {"$exists": "true"}}).limit(self.limit):
            rid = item["_id"]
            rdate = datetime.strptime(item["DT"], "%Y%m%d") + timedelta(hours=8)
            body = item["BODY"]
            rpt_typ = item["RPT_TYP"]
            rpt_sub_typ = item["RPT_SUB_TYP"]
            sec_cd = item["SEC_CD"]
            yield rid, rdate, body, rpt_typ, rpt_sub_typ, sec_cd

    def _preprocess_report(self, rst):
        rid, rdate, body, rpt_typ, rpt_sub_typ, sec_cd = rst

        if self.debug_mode:
            print(rid, rdate, len(body))
        paragraphs = []
        for p, paragraph in enumerate(body):
            sentences = []
            # print('-----------------')
            for i, sentence in enumerate(paragraph):
                sentence = re.sub(" +", " ", sentence).strip()

                pattern = "[가-힣A-Za-z\d'\"]"
                pos = re.search(pattern, sentence)
                if pos:
                    sentence = sentence[pos.start() :].strip()
                    if len(sentence) > 0:
                        pattern = "(당사는.*(유동성\s*공급자|위탁\s*증권사|동\s*자료)|^(본|이)\s*(조사자료|보고서|리포트)(는|의))"
                        pos = re.search(pattern, sentence)
                        if pos:
                            if self.debug_mode:
                                print(sentence)
                                print("----------------------------skipping sentences")
                            break
                        if sentence not in sentences:
                            sentences.append(sentence)
            if len(sentences) > 0:
                paragraphs.append(sentences)

        rid = "r" + str(rid)

        if len(paragraphs) < 1:
            if self.debug_mode:
                print(" ### skipping a report with not enough number of sentences...")
            return None

        nos = sum([len(p) for p in paragraphs])
        if nos < 4:
            if self.debug_mode:
                print(" ### skipping a report with not enough number of sentences...")
            return None
        # if debug:
        #     print(paragraphs)
        text = "\n\n".join(["\n".join(p) for p in paragraphs])
        if self.debug_mode:
            print(text)

        doc = {
            "rid": rid,
            "rdate": rdate,
            "type": rpt_typ,
            "subtype": rpt_sub_typ,
            "nos": nos,
            "text": text,
        }
        return doc



import logging
import pandas as pd
from ekorpkit import eKonf
from sklearn.model_selection import train_test_split
from sklearn import preprocessing
from ekorpkit.models.snorkel.labeling import (
    LabelingFunction,
    PandasLFApplier,
    LFAnalysis,
)
from ekorpkit.models.snorkel.labeling.model import LabelModel
from ekorpkit.models.snorkel.analysis import metric_score


log = logging.getLogger(__name__)


class BaseSnorkel:
    ABSTAIN = -1

    def __init__(self, data=None, **args):
        self._data = data
        args = eKonf.to_config(args)
        self.args = args
        self.verbose = args.get("verbose", False)
        self._path = args.path
        self._columns = args.columns
        self.ABSTAIN = args.get("ABSTAIN", -1)
        self._worker_lfs = None
        self._le = None
        self._classes = None
        self._train_data = None
        self._test_data = None
        self.L_train = None
        self.L_test = None
        self.label_model = None
        self.Y_train = None
        self.Y_test = None
        self.applier = None
        self._preds = None

    def load_data(self, data=None, test_size=0.2, random_state=12345, shuffle=True):
        if data is None:
            data = self._data

        if isinstance(data, pd.DataFrame):
            _data = data.copy()
        else:
            if eKonf.is_instantiatable(data):
                data = eKonf.instantiate(data)
            _data = data.data.copy()

        if _data is None:
            raise ValueError("No data to process")

        _columns = self.columns
        le = preprocessing.LabelEncoder()
        _labels = _data[_columns.labels]
        le.fit(_labels)
        _data[_columns.classes] = le.transform(_labels)

        _train_data, _test_data = train_test_split(
            _data, test_size=test_size, random_state=random_state, shuffle=shuffle
        )
        log.info(f"Train data: {_train_data.shape}, Test data: {_test_data.shape}")
        if self.verbose:
            print("Train data:", _train_data.shape)
            print(_train_data.head())
            print("Test data:", _test_data.shape)
            print(_test_data.head())

        self._le = le
        self._classes = le.classes_
        self._data = _data
        self._train_data = _train_data
        self._test_data = _test_data

    @property
    def columns(self):
        return self._columns

    @property
    def data(self):
        return self._data

    @property
    def classes(self):
        return self._classes

    @property
    def train_data(self):
        return self._train_data

    @property
    def test_data(self):
        return self._test_data

    def compose_worker_lfs(self):
        labels_by_annotator = self.data.groupby(self.columns.annotator)
        worker_dicts = {}
        for worker_id in labels_by_annotator.groups:
            worker_df = labels_by_annotator.get_group(worker_id)
            worker_dicts[worker_id] = dict(zip(worker_df.id, worker_df.classes))

        log.info(f"Number of workers: {len(worker_dicts)}")

        def worker_lf(x, worker_dict):
            return worker_dict.get(x.id, self.ABSTAIN)

        def make_worker_lf(worker_id):
            worker_dict = worker_dicts[worker_id]
            name = f"worker_{worker_id}"
            return LabelingFunction(
                name, f=worker_lf, resources={"worker_dict": worker_dict}
            )

        worker_lfs = [make_worker_lf(worker_id) for worker_id in worker_dicts]
        self._worker_lfs = worker_lfs
        return worker_lfs

    def apply_worker_lfs(self, worker_lfs=None):
        if worker_lfs is None:
            worker_lfs = self._worker_lfs
        applier = PandasLFApplier(worker_lfs)
        self.applier = applier
        log.info("Applying worker lfs to train data")
        self.L_train = applier.apply(self.train_data)
        log.info("Applying worker lfs to test data")
        self.L_test = applier.apply(self.test_data)
        self.Y_train = self.train_data[self.columns.classes].values
        self.Y_test = self.test_data[self.columns.classes].values

    def lf_summary(self, worker_lfs=None, L_train=None, Y_train=None):
        L_train = L_train or self.L_train
        Y_train = Y_train or self.Y_train

        if worker_lfs is None:
            worker_lfs = self._worker_lfs
        log.info(
            f"Training set coverage: {100 * LFAnalysis(L_train).label_coverage(): 0.1f}%"
        )
        return LFAnalysis(L_train, worker_lfs).lf_summary(Y_train)

    def fit(
        self,
        L_train=None,
        cardinality=None,
        verbose=False,
        n_epochs=100,
        seed=12345,
        log_freq=20,
        l2=0.1,
        lr=0.01,
        progress_bar=True,
        **args,
    ):
        L_train = L_train or self.L_train
        if cardinality is None:
            cardinality = len(self.classes)
        # Train LabelModel.
        self.label_model = LabelModel(cardinality=cardinality, verbose=verbose)
        self.label_model.fit(
            L_train,
            n_epochs=n_epochs,
            seed=seed,
            log_freq=log_freq,
            l2=l2,
            lr=lr,
            progress_bar=progress_bar,
        )

    def eval(
        self,
        L_train=None,
        Y_train=None,
        L_test=None,
        Y_test=None,
        tie_break_policy="abstain",
        metric="accuracy",
    ):
        L_test = L_test or self.L_test
        Y_test = Y_test or self.Y_test
        L_train = L_train or self.L_train
        Y_train = Y_train or self.Y_train

        preds_train = self.label_model.predict(
            L_train, tie_break_policy=tie_break_policy
        )
        acc = metric_score(Y_train, preds_train, probs=None, metric=metric)
        print(f"LabelModel Accuracy for train: {acc:.3f}")

        preds_test = self.label_model.predict(L_test, tie_break_policy=tie_break_policy)
        acc = metric_score(Y_test, preds_test, probs=None, metric=metric)
        print(f"LabelModel Accuracy for test: {acc:.3f}")

    def predict(self, data=None, columns=None, tie_break_policy="abstain"):
        data = data or self.data.copy()
        columns = columns or self.columns
        L_data = self.applier.apply(data)

        _snorkel_classes = columns["snorkel_classes"]
        _classes = columns["classes"]
        _labels = columns["labels"]

        data[_snorkel_classes] = self.label_model.predict(
            L_data, tie_break_policy=tie_break_policy
        )
        data = data[data.snorkel_classes != self.ABSTAIN]

        data[_labels] = self._le.inverse_transform(data[_snorkel_classes])
        self._preds = data
        if self.verbose:
            print("Data that predictions are different from classes:")
            print(data.query(f"{_classes} != {_snorkel_classes}"))
        return data

    def save_preds(self, data, columns=None, **kwargs):
        if data is None:
            data = self._preds
        if data is None:
            raise ValueError("No predictions to save")
        if isinstance(columns, list):
            data = data[columns].drop_duplicates()

        args = eKonf.to_dict(eKonf.merge(self._path.output, kwargs))
        eKonf.save_data(data, **args)



import numpy as np
import pandas as pd

from .apply.core import PandasSFApplier
from .sf import SlicingFunction


def slice_dataframe(
    df: pd.DataFrame, slicing_function: SlicingFunction
) -> pd.DataFrame:
    """Return a dataframe with examples corresponding to specified ``SlicingFunction``.

    Parameters
    ----------
    df
        A pandas DataFrame that will be sliced
    slicing_function
        SlicingFunction which will operate over df to return a subset of examples;
        function returns a subset of data for which ``slicing_function`` output is True

    Returns
    -------
    pd.DataFrame
        A DataFrame including only examples belonging to slice_name
    """

    S = PandasSFApplier([slicing_function]).apply(df)

    # Index into the SF labels by name
    df_idx = np.where(S[slicing_function.name])[0]  # type: ignore
    return df.iloc[df_idx]



import os
import logging
import hydra
from ekorpkit import eKonf, __hydra_version_base__
from ekorpkit.info.docs import make_table


log = logging.getLogger(__name__)


def cmd(**args):
    eKonf.run(args)


def listup(**args):
    ekorpkit_dir = eKonf.__ekorpkit_path__
    corpus_info = {}
    for name in args[eKonf.Keys.CORPUS]["preset"][eKonf.Keys.CORPUS]:
        info_path = f"{ekorpkit_dir}/resources/corpora/{name}.yaml"
        info = eKonf.load(info_path)
        corpus_info[name] = info
    make_table(corpus_info.values(), args["info"]["table"])


def about(**args):
    from . import __version__

    name = args.get("name")
    print()
    for k, v in args.items():
        print(f"{k:11} : {v}")
    print(f"\nExecute `{name} --help` to see what eKorpkit provides")


@hydra.main(config_path="conf", config_name="config", version_base=__hydra_version_base__)
def hydra_main(cfg) -> None:
    verbose = cfg.verbose

    if verbose:
        log.info("## eKorpkit Command Line Interface for Hydra ##")
    eKonf._init_env_(cfg, verbose)

    if cfg.get("print_config"):
        if cfg.get("print_resolved_config"):
            log.info("## hydra configuration resolved ##")
            eKonf.pprint(cfg)
        else:
            log.info("## hydra configuration ##")
            print(eKonf.to_yaml(cfg))

    if verbose:
        log.info(f"Hydra working directory : {os.getcwd()}")
        log.info(f"Orig working directory  : {hydra.utils.get_original_cwd()}")

    eKonf.instantiate(cfg)

    eKonf._stop_env_(cfg, verbose)


if __name__ == "__main__":
    hydra_main()



from typing import List


class Policy:
    """Base class for augmentation policies.

    Augmentation policies generate sequences of indices, corresponding
    to a sequence of transformation functions to be run on a data point.

    Parameters
    ----------
    n_tfs
        Total number of TFs
    n_per_original
        Number of transformed data points for each original data point
    keep_original
        Keep untransformed data point in augmented data set? Note that
        even if in-place modifications are made to the original data
        point by the TFs being applied, the original data point will
        remain unchanged.

    Raises
    ------
    NotImplementedError
        Subclasses need to implement the ``generate`` method

    Attributes
    ----------
    n
        Total number of TFs
    n_per_original
        See above
    keep_original
        See above
    """

    def __init__(
        self, n_tfs: int, n_per_original: int = 1, keep_original: bool = True
    ) -> None:
        self.n = n_tfs
        self.n_per_original = n_per_original
        self.keep_original = keep_original

    def generate_for_example(self) -> List[List[int]]:
        """Generate all sequences of TF indices for a single example.

        Generates `n_per_original` sequences, and adds an empty
        sequence if `keep_original` is True.

        Returns
        -------
        List[List[int]]
            Sequences of indices of TFs to run on data point in order.
        """
        seqs: List[List[int]] = []
        if self.keep_original:
            # Keep original by adding a null transformation sequence
            seqs.append([])
        for _ in range(self.n_per_original):
            seqs.append(self.generate())
        return seqs

    def generate(self) -> List[int]:
        """Generate a sequence of TF indices.

        Returns
        -------
        List[int]
            Indices of TFs to run on data point in order.

        Raises
        ------
        NotImplementedError
            Subclasses need to implement this method
        """
        raise NotImplementedError


class ApplyAllPolicy(Policy):
    """Apply all TFs in order to each data point.

    While this can be used as a baseline policy, using a
    random policy is more standard. See ``RandomPolicy``.

    Parameters
    ----------
    n_tfs
        Total number of TFs
    n_per_original
        Number of transformed data points for each original data point
    keep_original
        Keep untransformed data point in augmented data set? Note that
        even if in-place modifications are made to the original data
        point by the TFs being applied, the original data point will
        remain unchanged.

    Example
    -------
    >>> policy = ApplyAllPolicy(3, n_per_original=2, keep_original=False)
    >>> policy.generate_for_example()
    [[0, 1, 2], [0, 1, 2]]

    Attributes
    ----------
    n
        Total number of TFs
    n_per_original
        See above
    keep_original
        See above
    """

    def generate(self) -> List[int]:
        """Generate indices of all TFs in order.

        Returns
        -------
        List[int]
            Indices of all TFs in order.
        """
        return list(range(self.n))


class ApplyOnePolicy(ApplyAllPolicy):
    """Apply a single TF to each data point."""

    def __init__(self, n_per_original: int = 1, keep_original: bool = True) -> None:
        super().__init__(
            n_tfs=1, n_per_original=n_per_original, keep_original=keep_original
        )


class ApplyEachPolicy(Policy):
    """Apply each TF individually to each data point.

    This can be used as a baseline policy when using
    complex transformations which might degenerate if combined.

    Parameters
    ----------
    n_tfs
        Total number of TFs
    keep_original
        Keep untransformed data point in augmented data set? Note that
        even if in-place modifications are made to the original data
        point by the TFs being applied, the original data point will
        remain unchanged.

    Example
    -------
    >>> policy = ApplyEachPolicy(3, keep_original=True)
    >>> policy.generate_for_example()
    [[], [0], [1], [2]]

    Attributes
    ----------
    n
        Total number of TFs
    n_per_original
        Total number of TFs
    keep_original
        See above
    """

    def __init__(self, n_tfs: int, keep_original: bool = True) -> None:
        super().__init__(n_tfs=n_tfs, n_per_original=n_tfs, keep_original=keep_original)

    def generate_for_example(self) -> List[List[int]]:
        """Generate all length-one sequences for a single example.

        Returns
        -------
        List[List[int]]
            Sequences of indices of TFs to run on data point in order.
        """
        seqs = [[i] for i in range(self.n)]
        if self.keep_original:
            # Keep original by adding a null transformation sequence
            seqs.insert(0, [])
        return seqs



import functools
import tensorflow as tf

# import tensorflow.compat.v1 as tf
import tensorflow_datasets as tfds


def dumping_dataset(split, shuffle_files=False):
    del shuffle_files
    if split == "train":
        ds = tf.data.TextLineDataset(
            [
                "gs://scifive/finetune/BC5CDR-chem/train.tsv_cleaned.tsv",
            ]
        )
    else:
        ds = tf.data.TextLineDataset(
            [
                "gs://scifive/finetune/BC5CDR-chem/test.tsv_cleaned.tsv",
            ]
        )
    # Split each "<t1>\t<t2>" example into (input), target) tuple.
    ds = ds.map(
        functools.partial(
            tf.io.decode_csv,
            record_defaults=["", ""],
            field_delim="\t",
            use_quote_delim=False,
        ),
        num_parallel_calls=tf.data.experimental.AUTOTUNE,
    )
    # Map each tuple to a {"input": ... "target": ...} dict.
    ds = ds.map(lambda *ex: dict(zip(["input", "target"], ex)))
    return ds


def ner_preprocessor(ds):
    def normalize_text(text):
        """Lowercase and remove quotes from a TensorFlow string."""
        return text

    def to_inputs_and_targets(ex):
        """Map {"inputs": ..., "targets": ...}->{"inputs": ner..., "targets": ...}."""
        return {
            "inputs": tf.strings.join(
                ["bc5cdr_chem_ner: ", normalize_text(ex["input"])]
            ),
            "targets": normalize_text(ex["target"]),
        }

    return ds.map(
        to_inputs_and_targets, num_parallel_calls=tf.data.experimental.AUTOTUNE
    )


print("A few raw validation examples...")
for ex in tfds.as_numpy(dumping_dataset("train").take(5)):
    print(ex)

ds = dumping_dataset("train")
df = tfds.as_dataframe(ds)
df.tail()


# t5.data.TaskRegistry.remove('ncbi_ner')
# t5.data.TaskRegistry.add(
#     "ncbi_ner",
#     # Supply a function which returns a tf.data.Dataset.
#     dataset_fn=dumping_dataset,
#     splits=["train", "validation"],
#     # Supply a function which preprocesses text from the tf.data.Dataset.
#     text_preprocessor=[ner_preprocessor],
#     # Lowercase targets before computing metrics.
#     postprocess_fn=t5.data.postprocessors.lower_text,
#     # We'll use accuracy as our evaluation metric.
#     metric_fns=[t5.evaluation.metrics.accuracy,
#                t5.evaluation.metrics.sequence_accuracy,
#                 ],
#     output_features=t5.data.Feature(vocabulary=t5.data.SentencePieceVocabulary(vocab)),
#     # output_features=t5.data.Feature(vocabulary=t5.data.SentencePieceVocabulary(vocab))
# )



import tarfile
import orjson as json
from pathlib import Path
from bs4 import BeautifulSoup
import multiprocessing as mp


class GlassDoor:
    def __init__(
        self,
        name,
        input_file,
        output_dir,
        output_file,
        file_extension,
        num_workers,
        force_download,
        **kwargs,
    ):

        output_dir = Path(output_dir)
        if not output_dir.is_dir():
            output_dir.mkdir(exist_ok=True)
        output_path = output_dir / output_file

        if output_path.is_file() and not force_download:
            print(f"{output_path} already exists, skipping")
            return

        with tarfile.open(input_file) as tar:
            files = [file for file in tar.getmembers() if file_extension in file.name]
            print(f"Number of files in [{input_file}]: {len(files)}")
            processes = num_workers if num_workers else 1
            pool = mp.Pool(processes=processes)
            results = []
            for fno, file in enumerate(files):
                print("==> processing {}th file: {}".format(fno + 1, file.name))
                f = tar.extractfile(file)
                if f:
                    data = f.read()
                    results.append(pool.apply_async(parse_gd_review_html, args=(data,)))
            pool.close()
            pool.join()
            with open(output_path, "w") as f:
                for result in results:
                    reviews = result.get()
                    for line in reviews:
                        json.dump(line, f)
                        f.write("\n")

        print(f"Corpus [{name}] is built to [{output_path}] from [{input_file}]")


def parse_gd_review_html(data):
    soup = BeautifulSoup(data, "html.parser")
    reviews = []
    rvs = soup.find_all("li")
    for i, rv in enumerate(rvs):
        if rv.get("id") and rv.get("id").startswith("empReview_"):
            # print(f'parsing review {i}')
            review_dict = parse_gd_review(rv)
            reviews.append(review_dict)
            # print(review_dict)
    return reviews


def parse_gd_review(review):
    review_dict = {}

    rv_id = review.get("id")
    review_dict["id"] = rv_id
    rv_date = review.find("time", class_="date subtle small")
    rv_date = rv_date.get("datetime") if rv_date else None
    review_dict["date"] = rv_date
    rv_title = review.find("h2", class_="h2 summary strong mt-0 mb-xsm")
    if rv_title:
        rv_link = rv_title.find("a")
        if rv_link:
            rv_link = rv_title.find("a").get("href")
        rv_title = rv_title.get_text().replace('"', "")
    review_dict["link"] = rv_link
    review_dict["title"] = rv_title

    author = review.find("span", class_="authorJobTitle middle reviewer")
    author = author.get_text() if author else None
    review_dict["author"] = author

    ratings = {}
    srating = review.find(
        "div", class_="subRatings module stars__StarsStyles__subRatings"
    )
    if srating:
        for li in srating.find_all("li"):
            cat = li.find("div").get_text()
            rating = li.find("span", class_="gdBars gdRatings med").get("title")
            ratings[cat] = rating

    review_dict["ratings"] = ratings

    reviews = {}
    review_txt = ""
    rvtexts = review.find_all(
        "div", class_="mt-md common__EiReviewTextStyles__allowLineBreaks"
    )
    if rvtexts:
        for rvtext in rvtexts:
            text = ""
            cat = ""
            for p in rvtext.find_all("p"):
                if p.get("class"):
                    if p.get("class") == ["strong"]:
                        cat = p.get_text()
                else:
                    text += p.get_text().replace("\r", " ").strip() + "\n"
            if cat:
                reviews[cat] = text.strip()
            review_txt += text + "\n"
    review_dict["reviews"] = reviews
    review_dict["text"] = review_txt.strip()

    return review_dict



from typing import Any

import numpy as np

from .base_labeler import BaseLabeler


class RandomVoter(BaseLabeler):
    """Random vote label model.

    Example
    -------
    >>> L = np.array([[0, 0, -1], [-1, 0, 1], [1, -1, 0]])
    >>> random_voter = RandomVoter()
    >>> predictions = random_voter.predict_proba(L)
    """

    def predict_proba(self, L: np.ndarray) -> np.ndarray:
        """
        Assign random votes to the data points.

        Parameters
        ----------
        L
            An [n, m] matrix of labels

        Returns
        -------
        np.ndarray
            A [n, k] array of probabilistic labels

        Example
        -------
        >>> L = np.array([[0, 0, -1], [-1, 0, 1], [1, -1, 0]])
        >>> random_voter = RandomVoter()
        >>> predictions = random_voter.predict_proba(L)
        """
        n = L.shape[0]
        Y_p = np.random.rand(n, self.cardinality)
        Y_p /= Y_p.sum(axis=1).reshape(-1, 1)
        return Y_p


class MajorityClassVoter(BaseLabeler):
    """Majority class label model."""

    def fit(  # type: ignore
        self, balance: np.ndarray, *args: Any, **kwargs: Any
    ) -> None:
        """Train majority class model.

        Set class balance for majority class label model.

        Parameters
        ----------
        balance
            A [k] array of class probabilities
        """
        self.balance = balance

    def predict_proba(self, L: np.ndarray) -> np.ndarray:
        """Predict probabilities using majority class.

        Assign majority class vote to each datapoint.
        In case of multiple majority classes, assign equal probabilities among them.


        Parameters
        ----------
        L
            An [n, m] matrix of labels

        Returns
        -------
        np.ndarray
            A [n, k] array of probabilistic labels

        Example
        -------
        >>> L = np.array([[0, 0, -1], [-1, 0, 1], [1, -1, 0]])
        >>> maj_class_voter = MajorityClassVoter()
        >>> maj_class_voter.fit(balance=np.array([0.8, 0.2]))
        >>> maj_class_voter.predict_proba(L)
        array([[1., 0.],
               [1., 0.],
               [1., 0.]])
        """
        n = L.shape[0]
        Y_p = np.zeros((n, self.cardinality))
        max_classes = np.where(self.balance == max(self.balance))
        for c in max_classes:
            Y_p[:, c] = 1.0
        Y_p /= Y_p.sum(axis=1).reshape(-1, 1)
        return Y_p


class MajorityLabelVoter(BaseLabeler):
    """Majority vote label model."""

    def predict_proba(self, L: np.ndarray) -> np.ndarray:
        """Predict probabilities using majority vote.

        Assign vote by calculating majority vote across all labeling functions.
        In case of ties, non-integer probabilities are possible.

        Parameters
        ----------
        L
            An [n, m] matrix of labels

        Returns
        -------
        np.ndarray
            A [n, k] array of probabilistic labels

        Example
        -------
        >>> L = np.array([[0, 0, -1], [-1, 0, 1], [1, -1, 0]])
        >>> maj_voter = MajorityLabelVoter()
        >>> maj_voter.predict_proba(L)
        array([[1. , 0. ],
               [0.5, 0.5],
               [0.5, 0.5]])
        """
        n, m = L.shape
        Y_p = np.zeros((n, self.cardinality))
        for i in range(n):
            counts = np.zeros(self.cardinality)
            for j in range(m):
                if L[i, j] != -1:
                    counts[L[i, j]] += 1
            Y_p[i, :] = np.where(counts == max(counts), 1, 0)
        Y_p /= Y_p.sum(axis=1).reshape(-1, 1)
        return Y_p



import subprocess
import omegaconf
from hydra.utils import instantiate
from pathlib import Path


def _load_dataset(
    data_files,
    extension="json",
    test_split_percentage=10,
    min_tokens=5,
    download_mode="reuse_dataset_if_exists",
    num_proc=1,
):
    from datasets import load_dataset

    dataset = {}
    if test_split_percentage > 0:
        dataset["test"] = load_dataset(
            extension,
            data_files=data_files,
            split=f"train[:{test_split_percentage}%]",
            download_mode=download_mode,
        )
        dataset["train"] = load_dataset(
            extension,
            data_files=data_files,
            split=f"train[{test_split_percentage}%:]",
        )
    else:
        dataset = load_dataset(
            extension, data_files=data_files, download_mode=download_mode
        )
    for name in dataset:
        n_proc = min(num_proc, len(dataset[name]) - 1)
        cols = dataset[name].column_names
        cols.remove("text")
        if len(cols) > 0:
            dataset[name] = dataset[name].remove_columns(cols)
        dataset[name] = dataset[name].filter(
            lambda x: len(x["text"].split()) > min_tokens, num_proc=n_proc
        )
        # print(dataset[name]['text'][0])
        print(
            "Loading Datasets: There are",
            len(dataset[name]),
            f"records in {name} dataset.",
            data_files,
        )
    return dataset


class ShardingDataset:
    def __init__(
        self, corpus, output, dataset, n_shards, n_processes, segmenter, **args
    ):

        self.input_extension = corpus["corpus_filetype"]
        corpus_files = corpus["corpus_files"]
        corpus_dir = corpus["corpus_dir"]
        if corpus_files:
            print(type(corpus_files))
            if isinstance(corpus_files, str):
                corpus_files = {corpus_files: 1}
            elif (
                isinstance(corpus_files, list)
                or type(corpus_files) == omegaconf.listconfig.ListConfig
            ):
                corpus_files = {f: 1 for f in corpus_files}
            else:
                corpus_files = dict(corpus_files)
            corpus_files = {corpus_dir + "/" + f: d for f, d in corpus_files.items()}
        else:
            corpus_files = {
                str(f): 1 for f in Path(corpus_dir).glob("**/*") if f.is_file()
            }
        assert (
            len(corpus_files) > 0
        ), "The input file list must contain at least one file."
        self.input_files = corpus_files

        assert n_shards["train"] > 0, "There must be at least one output shard."
        # assert n_test_shards > 0, 'There must be at least one output shard.'
        self.n_shards = n_shards

        self.download_mode = dataset["download_mode"]
        self.fraction_test_set = dataset["fraction_test_set"]
        if not self.fraction_test_set > 0:
            self.n_shards["test"] = 0
        self.shuffle_dataset = dataset["shuffle_dataset"]
        self.seed = dataset["seed"]
        self.min_tokens = dataset["min_tokens"]

        if n_processes:
            self.n_processes = n_processes
        else:
            self.n_processes = 7
        self.segmenter = segmenter

        self.output_dir = Path(output["output_dir"])
        if not self.output_dir.is_dir():
            self.output_dir.mkdir(parents=True, exist_ok=True)
        self.output_name_prefix = output["name_prefix"]
        self.output_identifier = output["identifier"]
        self.output_file_extension = output["file_extension"]

        self.datasets = {}  # key: split name, value: dataset
        self.output_files = {
            "train": {},
            "test": {},
        }  # key: filename, value: list of articles to go into file

        self.init_output_files()

    # Remember, the input files contain one article per line (the whitespace check is to skip extraneous blank lines)
    def load_datasets(self):
        from datasets import concatenate_datasets

        print("Start: Loading Datasets")

        datasets = {"train": [], "test": []}
        for input_file, dupe_factor in self.input_files.items():
            print(f"processing {input_file} with {dupe_factor} dupe factor")
            dset = _load_dataset(
                input_file,
                extension=self.input_extension,
                test_split_percentage=int(self.fraction_test_set * 100),
                min_tokens=self.min_tokens,
                download_mode=self.download_mode,
                num_proc=self.n_processes,
            )
            # self.datasets[input_file] = dset
            for name in dset:
                for i in range(dupe_factor):
                    datasets[name].append(dset[name])
                    print(f"{i}: adding {len(dset[name])} records in {name} dataset.")
        for name in datasets:
            if len(datasets[name]) > 0:
                self.datasets[name] = concatenate_datasets(datasets[name])
                print(f"Concatenating {len(datasets[name])} {name} datasets.")
        if self.shuffle_dataset:
            for name in self.datasets:
                print(f"Shuffling {name} dataset.")
                self.datasets[name] = self.datasets[name].shuffle(self.seed)

        print("End: Loading Datasets.")

    def segment_articles_into_sentences(self):
        if not self.datasets:
            self.load_articles()

        assert (
            len(self.datasets["train"]) > 0
        ), "Please check that input files are present and contain data."
        assert self.segmenter[
            "_target_"
        ], "Please check that the segmenter is configured correctly."
        print("Start: Sentence Segmentation")

        seg = instantiate(self.segmenter)

        def segment_into_sentences(examples):
            sentences = []
            for text in examples["text"]:
                sentences += seg.segment_article(text.replace("\t", " "))
                sentences.append("")
            return {"text": sentences}

        num_articles = 0
        num_sentences = 0
        for name in self.datasets:
            num_articles += len(self.datasets[name])
            num_proc = min(self.n_processes, len(self.datasets[name]) - 1)
            self.datasets[name] = self.datasets[name].map(
                segment_into_sentences, batched=True, num_proc=num_proc
            )
            num_sentences += len(self.datasets[name])

        print(
            f"Number of articles: {num_articles}, Number of sentences: {num_sentences}"
        )
        print("End: Sentence Segmentation")

    def init_output_files(self):
        print("Start: Init Output Files")
        for name in self.output_files:
            assert (
                len(self.output_files[name]) == 0
            ), "Internal storage self.output_files already contains data."

            for i in range(self.n_shards[name]):
                filename = (
                    self.output_name_prefix
                    + self.output_identifier[name]
                    + "_"
                    + str(i)
                    + self.output_file_extension
                )
                self.output_files[name][filename] = None

        print("End: Init Output Files")

    def distribute_datasets_over_shards(self):
        print("Start: Distribute Datasets Over Shards")
        for name in self.datasets:
            assert (
                len(self.datasets[name]) > self.n_shards[name]
            ), "There are fewer articles than shards. Please add more data or reduce the number of shards requested."

            for i, filename in enumerate(self.output_files[name]):
                self.output_files[name][filename] = self.datasets[name].shard(
                    self.n_shards[name], i, contiguous=True
                )

            for shard in self.output_files[name]:
                print(f"{name} shard:", shard, len(self.output_files[name][shard]))

        print("End: Distribute Datasets Over Shards")

    def write_shards_to_disk(self):
        print("Start: Write Shards to Disk")
        for name in self.output_files:
            for shard in self.output_files[name]:
                self.write_single_shard(shard, self.output_files[name][shard])

        print("End: Write Shards to Disk")
        for split, n in self.n_shards.items():
            if n > 0:
                if not (self.output_dir / split).is_dir():
                    (self.output_dir / split).mkdir(exist_ok=True, parents=True)
                absolute_dir = str(self.output_dir)
                command = (
                    "mv "
                    + absolute_dir
                    + "/*"
                    + split
                    + "*.txt"
                    + " "
                    + absolute_dir
                    + "/"
                    + split
                )
                print(command)
                mv_process = subprocess.Popen(command, shell=True)

                mv_process.wait()

    def write_single_shard(self, shard_name, shard_dataset):
        with open(shard_name, mode="w", newline="\n") as f:
            for text in shard_dataset["text"]:
                f.write(text + "\n")



import os
import logging
from collections import namedtuple
from .score import NEG_INF
from .base import (
    _match_any_rules,
    _prepare_ngram_tuples,
    _remove_overlaps,
    _get_ngram_tuple,
    _get_ngram_str,
    _KEEP,
)
from ekorpkit import eKonf


log = logging.getLogger(__name__)


class Ngrams:
    def __init__(
        self,
        **args,
    ):
        args = eKonf.to_config(args)

        self.name = args.name
        self._data = args.data

        self._ngram = args.ngram
        self._ngramize = args.ngramize
        self._postag = args.postag
        self._scores = args.scores
        self._info = args.info
        self.auto = args.auto
        self.force = args.force

        assert type(self._ngram.max_n) == int

        if self._ngram.max_n <= 0:
            self._ngram.max_n = 4
        if self._ngram.max_window < self._ngram.max_n:
            self._ngram.max_window = self._ngram.max_n
        if not self._ngram.max_skip:
            self._ngram.max_skip = self._ngram.max_window - self._ngram.max_n
        self.postag_rules = eKonf.ensure_list(self._ngram.postag_rules)

        if self.tokenize is None:
            self.tokenize = lambda x: x.split()
        self.progress_per = args.progress_per
        self.verbose = args.verbose

        self.score_path = args.scores.score_path

        self._tokenizer = args.preprocessor.tokenizer
        if eKonf.is_instantiatable(self._tokenizer):
            self._tokenizer = eKonf.instantiate(self._tokenizer)
        self._postag.stop_tags = eKonf.ensure_list(self._postag.stop_tags)

        self._sentences = []
        self.candidates = {}
        self._surface_to_tuples = {}

        if self.auto.load:
            eKonf.methods(args._method_, self)

    def __len__(self):
        return len(self.candidates)

    def __contains__(self, ngram):
        return ngram in self.candidates

    def __iter__(self):
        return iter(self.candidates)

    def __str__(self):
        classname = self.__class__.__name__
        s = f"{classname}\n----------\n"
        s += f"name: {self.name}\n"
        s += f"fullname: {self._info.fullname}\n"
        s += f"source: {self._info.source}\n"
        s += f"lang: {self._info.lang}\n"
        s += f"features: {self._score.features}\n"
        s += f"score path: {self.score_path}\n"
        s += f"no. of candidates: {len(self.candidates)}\n"
        return s

    def initialize(self):
        self.load_candidates()
        if not self.candidates or self.force.train:
            self.train()
            self.save_candidates()

    def tokenize(self, text):
        tokens = self._tokenizer.tokenize(text)
        return self._tokenizer.extract(
            tokens,
            postags=self._postag.use_tags,
            strip_pos=self._postag.strip_pos,
            stop_postags=self._postag.stop_tags,
            postag_length=self._postag.max_len,
        )

    def load_candidates(self):
        """Load a previously saved model"""
        if os.path.exists(self.score_path):
            _words = self._scores.columns.words
            _features = self._scores.features
            _lowercase = self._scores.lowercase

            df = eKonf.load_data(self.score_path, verbose=self.verbose)
            df = df.dropna(subset=[_words])
            columns = [_words] + _features
            Score = namedtuple("score", columns)
            _cands = df[columns].to_dict(orient="records")
            self.candidates = {}
            self._surface_to_tuples = {}
            for cand in _cands:
                ngram_tuple = self.to_ngram_tuple(
                    cand[_words],
                    ngram_delim=self._scores.ngram_delim,
                )
                ngram_str = self.to_ngram_str(
                    ngram_tuple,
                    ngram_delim=self._ngram.delimiter,
                    postag_delim=self._postag.delimiter,
                    strip_pos=self._postag.strip_pos,
                    postag_length=self._postag.max_len,
                    lowercase=_lowercase,
                )
                surface_str = self.to_ngram_str(
                    ngram_tuple,
                    ngram_delim="",
                    postag_delim=self._postag.delimiter,
                    strip_pos=True,
                    lowercase=_lowercase,
                )
                cand[_words] = ngram_str
                ngram_tuple = self.to_ngram_tuple(
                    ngram_str, ngram_delim=self._ngram.delimiter
                )
                self.candidates[ngram_tuple] = Score(**cand)
                self._surface_to_tuples[surface_str] = ngram_tuple

            log.info(f"loaded {len(self.candidates)} candidates")
        else:
            log.info("no candidates to load")

    def save_candidates(self):
        raise NotImplementedError

    @property
    def sentences(self):
        if len(self._sentences) == 0:
            self.load_data()
        return self._sentences

    def load_data(self):
        """Load data"""
        if self._data is None:
            log.warning("No data config found")
            return
        data = eKonf.instantiate(self._data)
        docs = data.data[data.COLUMN.TEXT]
        self._sentences = []
        for doc in docs:
            self._sentences.extend(doc.split("\n"))

    def train(self):
        """Train the model"""
        raise NotImplementedError

    def score_ngrams(self):
        raise NotImplementedError

    def learn_ngrams(self):
        raise NotImplementedError

    def find_features(self, words, features=None, **kwargs):
        """Find features for a given ngram tuple"""
        _features = features or self._scores.features
        return self.find_ngrams(
            words,
            features=_features,
            exclude_overlaps=self._ngramize.exclude_overlaps,
            overlaps_to_keep=self._ngramize.overlaps_to_keep,
            threshold=self._ngramize.threshold,
            ignore_scores=self._ngramize.ignore_scores,
            apply_postag_rules=self._ngramize.apply_postag_rules,
            use_surfaces_to_score=self._ngramize.use_surfaces_to_score,
            strip_pos=self._ngramize.strip_pos,
            surface_delim=self._ngramize.delimiter,
            postag_delim=self._postag.delimiter,
            postag_length=self._postag.max_len,
            **kwargs,
        )

    def find_ngrams(
        self,
        words,
        exclude_overlaps=None,
        overlaps_to_keep: _KEEP = None,
        threshold=None,
        ignore_scores=None,
        apply_postag_rules=None,
        use_surfaces_to_score=None,
        strip_pos=None,
        surface_delim=None,
        postag_delim=None,
        postag_length=None,
        features=None,
    ):
        exclude_overlaps = (
            exclude_overlaps
            if exclude_overlaps is not None
            else self._ngramize.exclude_overlaps
        )
        overlaps_to_keep = overlaps_to_keep or _KEEP(self._ngramize.overlaps_to_keep)
        threshold = threshold or self._ngramize.threshold
        ignore_scores = (
            ignore_scores if ignore_scores is not None else self._ngramize.ignore_scores
        )
        apply_postag_rules = (
            apply_postag_rules
            if apply_postag_rules is not None
            else self._ngramize.apply_postag_rules
        )
        use_surfaces_to_score = (
            use_surfaces_to_score
            if use_surfaces_to_score is not None
            else self._ngramize.use_surfaces_to_score
        )
        surface_delim = surface_delim or self._ngramize.delimiter
        postag_delim = postag_delim or self._postag.delimiter
        strip_pos = strip_pos if strip_pos is not None else self._ngramize.strip_pos
        postag_length = postag_length or self._postag.max_len

        postag_rules = self.postag_rules if apply_postag_rules else []

        _count = self._scores.columns.count
        _score = self._scores.columns.score
        features = features or [_score]

        results = {}
        for ngram, _, score in self.analyze_sentence(
            words,
            exclude_overlaps=exclude_overlaps,
            overlaps_to_keep=overlaps_to_keep,
            threshold=threshold,
            ignore_scores=ignore_scores,
            use_surfaces_to_score=use_surfaces_to_score,
            postag_rules=postag_rules,
        ):
            ngram_str = self.to_ngram_str(
                ngram,
                ngram_delim=surface_delim,
                strip_pos=strip_pos,
                postag_delim=postag_delim,
                postag_length=postag_length,
            )
            if ignore_scores or score > NEG_INF:
                if ngram_str in results:
                    results[ngram_str][_count] += 1
                else:
                    feats = self.get_features(ngram, features)
                    if feats:
                        feats[_count] = 1
                        results[ngram_str] = feats
        if self.verbose:
            log.info(f"found {len(results)} ngrams")
        return results

    def get_features(self, ngram, features):
        """Get features"""
        _features = {}
        score = self.candidates.get(ngram)
        if score is None or features is None:
            return _features
        for _f in features:
            if _f in score._fields:
                _features[_f] = getattr(score, _f)
        return _features

    def __getitem__(self, words):
        return self.ngramize_sentence(words)

    def ngramize_sentence(
        self,
        words,
        exclude_overlaps=None,
        overlaps_to_keep: _KEEP = None,
        threshold=None,
        ignore_scores=None,
        apply_postag_rules=None,
        use_surfaces_to_score=None,
        strip_pos=None,
        surface_delim=None,
        postag_delim=None,
        postag_length=None,
        **kwargs,
    ):
        """
        Return a list of ngrams of the sentence
        """
        exclude_overlaps = (
            exclude_overlaps
            if exclude_overlaps is not None
            else self._ngramize.exclude_overlaps
        )
        overlaps_to_keep = overlaps_to_keep or _KEEP(self._ngramize.overlaps_to_keep)
        threshold = threshold or self._ngramize.threshold
        ignore_scores = (
            ignore_scores if ignore_scores is not None else self._ngramize.ignore_scores
        )
        apply_postag_rules = (
            apply_postag_rules
            if apply_postag_rules is not None
            else self._ngramize.apply_postag_rules
        )
        use_surfaces_to_score = (
            use_surfaces_to_score
            if use_surfaces_to_score is not None
            else self._ngramize.use_surfaces_to_score
        )
        surface_delim = surface_delim or self._ngramize.delimiter
        postag_delim = postag_delim or self._postag.delimiter
        strip_pos = strip_pos if strip_pos is not None else self._ngramize.strip_pos
        postag_length = postag_length or self._postag.max_len

        postag_rules = self.postag_rules if apply_postag_rules else []

        return [
            self.to_ngram_str(
                ngram,
                ngram_delim=surface_delim,
                postag_delim=postag_delim,
                strip_pos=strip_pos,
                postag_length=postag_length,
            )
            for ngram, _, _ in self.analyze_sentence(
                words,
                exclude_overlaps=exclude_overlaps,
                overlaps_to_keep=overlaps_to_keep,
                threshold=threshold,
                ignore_scores=ignore_scores,
                use_surfaces_to_score=use_surfaces_to_score,
                postag_rules=postag_rules,
            )
        ]

    def analyze_sentence(
        self,
        words,
        exclude_overlaps=True,
        overlaps_to_keep: _KEEP = _KEEP.HIGHEST_SCORE,
        threshold=None,
        ignore_scores=False,
        use_surfaces_to_score=False,
        postag_rules=[],
        **kwargs,
    ):
        """Analyze a sentence, concatenating any detected ngrams into a single token.

        Parameters
        ----------
        sentence : iterable of str
            Token sequence representing the sentence to be analyzed.
        """

        words = self.tokenize(words) if isinstance(words, str) else words
        ngram_with_positions = self.prepare_ngram_tuples(
            words,
            max_n=self._ngram.max_n,
            max_window=self._ngram.max_window,
            max_skip=self._ngram.max_skip,
            postag_rules=postag_rules,
            postag_delim=self._postag.delimiter,
            include_positions=True,
        )

        ngrams = []
        for ngram_pos in ngram_with_positions:
            ngram, pos = ngram_pos
            score = self.ngram_score(
                ngram, threshold=threshold, use_surfaces_to_score=use_surfaces_to_score
            )
            if score is not None or ignore_scores:
                ngrams.append((ngram, pos, score))
        if exclude_overlaps:
            ngrams = self.remove_overlaps(ngrams, keep=overlaps_to_keep)
        return ngrams

    def ngram_score(
        self,
        ngram,
        threshold=None,
        unigram_score=NEG_INF,
        use_surfaces_to_score=False,
    ):
        """Score a ngram"""

        _score = self._scores.columns.score

        if use_surfaces_to_score:
            surface_str = self.to_ngram_str(
                ngram,
                ngram_delim="",
                postag_delim=self._postag.delimiter,
                strip_pos=True,
            )
            if surface_str in self._surface_to_tuples:
                ngram = self._surface_to_tuples[surface_str]
                score = getattr(self.candidates[ngram], _score)
                if threshold is None or score >= threshold:
                    return score

        if ngram in self.candidates:
            score = getattr(self.candidates[ngram], _score)
            if threshold is None or score >= threshold:
                return score
        if len(ngram) == 1:
            return unigram_score
        return None

    def export_ngrams(
        self, threshold=None, postag_rules=None
    ):
        """Extract all found ngrams.
        Returns
        ------
        dict(str, float)
            Mapping between phrases and their scores.
        """
        postag_rules = postag_rules or self.postag_rules

        result = {}
        _score = self._scores.columns.score
        for ngram, score in self.candidates.items():
            if len(ngram) < 2:
                continue  # no phrases here
            if postag_rules and not self.match_any_rules(
                ngram, postag_rules, self._postag.delimiter
            ):
                continue
            if threshold is None or getattr(score, _score) > threshold:
                result[ngram] = score
        # sort by score
        result = sorted(
            result.items(), key=lambda x: getattr(x[1], _score), reverse=True
        )
        return result

    @staticmethod
    def match_any_rules(ngram, postag_rules, postag_delimiter):
        return _match_any_rules(ngram, postag_rules, postag_delimiter)

    @staticmethod
    def prepare_ngram_tuples(
        words,
        max_n=5,
        max_window=None,
        max_skip=None,
        postag_rules=[],
        postag_delim=";",
        include_positions=False,
    ):
        """Prepare ngram tuples from a list of words."""

        return _prepare_ngram_tuples(
            words,
            max_n=max_n,
            max_window=max_window,
            max_skip=max_skip,
            postag_rules=postag_rules,
            postag_delim=postag_delim,
            include_positions=include_positions,
        )

    @staticmethod
    def remove_overlaps(ngram_pos_scores, keep="highest_score"):
        """Remove overlapping ngrams by score.

        Parameters
        ----------
        ngram_pos_scores : list of (str, int, float)
            List of ngrams and their positions in the sentence.

        Returns
        -------
        list of (str, int, float)
            List of ngrams and their positions in the sentence.
        """
        return _remove_overlaps(ngram_pos_scores, keep=keep)

    @staticmethod
    def to_ngram_tuple(ngram_str, ngram_delim=";"):
        """Get a ngram tuple from a string."""
        return _get_ngram_tuple(ngram_str, ngram_delim)

    @staticmethod
    def to_ngram_str(
        ngram,
        ngram_delim="",
        postag_delim="/",
        strip_pos=True,
        postag_length=None,
        lowercase=True,
    ):
        """Get the surface form of a ngram tuple."""

        return _get_ngram_str(
            ngram,
            ngram_delim=ngram_delim,
            postag_delim=postag_delim,
            strip_pos=strip_pos,
            postag_length=postag_length,
            lowercase=lowercase,
        )



import os
import logging
import random
import numpy as np
import jax
import jax.numpy as jnp
from functools import partial
from glob import glob
from ekorpkit import eKonf
from ekorpkit.utils.func import elapsed_timer
from transformers import CLIPProcessor, FlaxCLIPModel
from flax.jax_utils import replicate
from flax.training.common_utils import shard_prng_key, shard
from PIL import Image
from tqdm.auto import trange
from .base import BaseTTIModel


log = logging.getLogger(__name__)


class DalleMini(BaseTTIModel):
    def __init__(self, **args):
        super().__init__(**args)

        self._num_devices = self.args.num_devices
        self.model = None
        self.model_params = None
        self.processor = None
        self.vqgan = None
        self.vqgan_params = None
        self.clip = None
        self.clip_params = None
        self.clip_processor = None
        self.devices = None

        if self.auto.load:
            self.load()

    def imagine(self, text_prompts=None, **args):

        """Diffuse the model"""

        log.info("> loading settings...")
        args = self.load_config(**args)

        text_prompts = text_prompts or eKonf.to_dict(args.text_prompts)
        if isinstance(text_prompts, str):
            text_prompts = [text_prompts]
        args.text_prompts = text_prompts

        self.sample_imagepaths = []

        tokenized_prompts = self.processor(args.text_prompts)
        print(f"Prompts: {args.text_prompts}\n")

        _batch_dir = self._output.batch_dir
        # _num_devices = min(self._num_devices, args.n_samples)
        _num_devices = self._num_devices
        _devices = self.devices[:_num_devices]
        log.info(f"Using {_num_devices} devices")
        log.info(f"Devices: {_devices}")
        args.num_samples = max(args.n_samples // _num_devices, 1) * _num_devices
        self._config = args
        self.save_settings(args)

        # Keys are passed to the model on each device to generate unique inference per device.
        # create a random key
        log.info(f"Seed used: {args.seed}")
        key = jax.random.PRNGKey(args.seed)
        # Model parameters are replicated on each device for faster inference.
        _model_params = replicate(self.model_params, devices=_devices)
        _vqgan_params = replicate(self.vqgan_params, devices=_devices)
        _tokenized_prompts = replicate(tokenized_prompts, devices=_devices)

        # Model functions are compiled and parallelized to take advantage of multiple devices.
        # model inference
        @partial(
            jax.pmap,
            axis_name="batch",
            static_broadcasted_argnums=(3, 4, 5, 6),
            devices=_devices,
        )
        def p_generate(
            tokenized_prompt, key, params, top_k, top_p, temperature, condition_scale
        ):
            return self.model.generate(
                **tokenized_prompt,
                prng_key=key,
                params=params,
                top_k=top_k,
                top_p=top_p,
                temperature=temperature,
                condition_scale=condition_scale,
            )

        # decode image
        @partial(jax.pmap, axis_name="batch", devices=_devices)
        def p_decode(indices, params):
            return self.vqgan.decode_code(indices, params=params)

        with elapsed_timer(format_time=True) as elapsed:

            # generate images
            # sample_images = []
            sample_num = 0
            for i in trange(max(args.n_samples // _num_devices, 1)):
                eKonf.clear_output(wait=True)
                # get a new key
                key, subkey = jax.random.split(key)
                # generate images
                encoded_images = p_generate(
                    _tokenized_prompts,
                    shard_prng_key(subkey),
                    _model_params,
                    args.gen_top_k,
                    args.gen_top_p,
                    args.temperature,
                    args.cond_scale,
                )
                # remove BOS
                encoded_images = encoded_images.sequences[..., 1:]
                # decode images
                decoded_images = p_decode(encoded_images, _vqgan_params)
                decoded_images = decoded_images.clip(0.0, 1.0).reshape(
                    (-1, 256, 256, 3)
                )
                for decoded_img in decoded_images:
                    img = Image.fromarray(np.asarray(decoded_img * 255, dtype=np.uint8))
                    eKonf.display(img)
                    # sample_images.append(img)
                    filename = (
                        f"{args.batch_name}({args.batch_num})_{sample_num:04}.png"
                    )
                    _img_path = os.path.join(_batch_dir, filename)
                    img.save(_img_path)
                    self.sample_imagepaths.append(_img_path)
                    log.info(f"Saved {filename}")
                    sample_num += 1

            eKonf.clear_output(wait=True)
            log.info(" >> elapsed time to diffuse: {}".format(elapsed()))
            print(f"{args.n_samples} samples generated to {_batch_dir}")
            print(f"text prompts: {text_prompts}")
            print("sample image paths:")
            for p in self.sample_imagepaths:
                print(p)

            if args.show_collage:
                self.collage(image_filepaths=self.sample_imagepaths)

    def load_models(self):
        from dalle_mini import DalleBart, DalleBartProcessor
        from vqgan_jax.modeling_flax_vqgan import VQModel

        # check how many devices are available
        log.info(f"Available devices: {jax.local_device_count()}")
        if self._num_devices:
            self._num_devices = min(self._num_devices, jax.local_device_count())
        else:
            self._num_devices = jax.local_device_count()
        self.devices = jax.local_devices()
        log.info(f"Using {self._num_devices} devices")
        log.info(f"Devices: {self.devices}")

        # Load dalle-mini
        self.model, self.model_params = DalleBart.from_pretrained(
            self.model_config.DALLE_MODEL,
            revision=self.model_config.DALLE_COMMIT_ID,
            dtype=jnp.float16,
            _do_init=self.model_config.DALLE_INIT,
        )
        self.processor = DalleBartProcessor.from_pretrained(
            self.model_config.DALLE_MODEL,
            revision=self.model_config.DALLE_COMMIT_ID,
        )
        # Load VQGAN
        self.vqgan, self.vqgan_params = VQModel.from_pretrained(
            self.model_config.VQGAN_REPO,
            revision=self.model_config.VQGAN_COMMIT_ID,
            _do_init=self.model_config.VQGAN_INIT,
        )

    def load_clip_models(self):
        self.clip, self.clip_params = FlaxCLIPModel.from_pretrained(
            self.model_config.CLIP_REPO,
            revision=self.model_config.CLIP_COMMIT_ID,
            dtype=jnp.float16,
            _do_init=self.model_config.CLIP_INIT,
        )
        self.clip_processor = CLIPProcessor.from_pretrained(
            self.model_config.CLIP_REPO,
            revision=self.model_config.CLIP_COMMIT_ID,
        )

    def rank_image_by_clip_score(self, prompts, images):
        """
        Rank an image by its clip score.
        """

        if self.clip is None:
            self.load_clip_models()

        _num_devices = self._num_devices
        _devices = self.devices[:_num_devices]
        log.info(f"Using {_num_devices} devices")
        log.info(f"Devices: {_devices}")

        _clip_params = replicate(self.clip_params, devices=_devices)

        # score images
        @partial(jax.pmap, axis_name="batch", devices=_devices)
        def p_clip(inputs, params):
            logits = self.clip(params=params, **inputs).logits_per_image
            return logits

        # from flax.training.common_utils import shard

        # get clip scores
        clip_inputs = self.clip_processor(
            text=prompts * _num_devices,
            images=images,
            return_tensors="np",
            padding="max_length",
            max_length=77,
            truncation=True,
        ).data
        logits = p_clip(shard(clip_inputs), _clip_params)

        # organize scores per prompt
        p = len(prompts)
        logits = np.asarray([logits[:, i::p, i] for i in range(p)]).squeeze()

        for i, prompt in enumerate(prompts):
            print(f"Prompt: {prompt}\n")
            for idx in logits[i].argsort()[::-1]:
                eKonf.display(images[idx * p + i])
                print(f"Score: {jnp.asarray(logits[i][idx], dtype=jnp.float32):.2f}\n")
            print()

    def load_config(self, batch_name=None, batch_num=None, **args):
        """Load the settings"""
        args = super().load_config(batch_name=batch_name, batch_num=batch_num, **args)

        if args.set_seed == "random_seed":
            random.seed()
            args.seed = random.randint(0, 2 ** 32 - 1)
        else:
            args.seed = int(args.set_seed)

        batch_arg_file = os.path.join(
            self._output.batch_dir, f"{args.batch_name}(*)_settings.yaml"
        )
        if args.resume_run:
            if args.run_to_resume == "latest":
                try:
                    args.batch_num
                except:
                    args.batch_num = len(glob(batch_arg_file)) - 1
            else:
                args.batch_num = int(args.run_to_resume)
        else:
            args.batch_num = len(glob(batch_arg_file))

        return args



import sys
import re
import requests
from bs4 import BeautifulSoup
from datetime import datetime

# Import parent class
from .base import FOMC


class Speech(FOMC):
    """
    A convenient class for extracting speech from the FOMC website
    """

    def __init__(self, content_type, **args):
        super().__init__(content_type, **args)
        self.speech_base_url = self.base_url + "/newsevents/speech"

    def _get_links(self, from_year):
        """
        Override private function that sets all the links for the contents to download on FOMC website
         from from_year (=min(2015, from_year)) to the current most recent year
        """
        self.links = []
        self.titles = []
        self.speakers = []
        self.dates = []

        res = requests.get(self.calendar_url)
        soup = BeautifulSoup(res.text, "html.parser")

        if self.verbose:
            print("Getting links for speeches...")
        to_year = datetime.today().strftime("%Y")

        if from_year <= 1995:
            print("Archive only from 1996, so setting from_year as 1996...")
            from_year = 1996
        for year in range(from_year, int(to_year) + 1):
            # Archived between 1996 and 2005, URL changed from 2011
            if year < 2011:
                speech_url = self.speech_base_url + "/" + str(year) + "speech.htm"
            else:
                speech_url = self.speech_base_url + "/" + str(year) + "-speeches.htm"

            res = requests.get(speech_url)
            soup = BeautifulSoup(res.text, "html.parser")
            speech_links = soup.findAll(
                "a",
                href=re.compile(
                    "^/?newsevents/speech/.*{}\d\d\d\d.*.htm|^/boarddocs/speeches/{}/|^{}\d\d\d\d.*.htm".format(
                        str(year), str(year), str(year)
                    )
                ),
            )
            for speech_link in speech_links:
                # Sometimes the same link is put for watch live video. Skip those.
                if speech_link.find({"class": "watchLive"}):
                    continue

                # Add link, title and date
                self.links.append(speech_link.attrs["href"])
                self.titles.append(speech_link.get_text())
                self.dates.append(
                    datetime.strptime(
                        self._date_from_link(speech_link.attrs["href"]), "%Y-%m-%d"
                    )
                )

                # Add speaker
                # Somehow the speaker is before the link in 1997 only, whereas the others is vice-versa
                if year == 1997:
                    # Somehow only the linke for December 15 speech has speader after the link in 1997 page.
                    if (
                        speech_link.get("href")
                        == "/boarddocs/speeches/1997/19971215.htm"
                    ):
                        tmp_speaker = (
                            speech_link.parent.next_sibling.next_element.get_text()
                            .replace("\n", "")
                            .strip()
                        )
                    else:
                        tmp_speaker = (
                            speech_link.parent.previous_sibling.previous_sibling.get_text()
                            .replace("\n", "")
                            .strip()
                        )
                else:
                    # Somehow 20051128 and 20051129 are structured differently
                    if speech_link.get("href") in (
                        "/boarddocs/speeches/2005/20051128/default.htm",
                        "/boarddocs/speeches/2005/20051129/default.htm",
                    ):
                        tmp_speaker = (
                            speech_link.parent.previous_sibling.previous_sibling.get_text()
                            .replace("\n", "")
                            .strip()
                        )
                    tmp_speaker = (
                        speech_link.parent.next_sibling.next_element.get_text()
                        .replace("\n", "")
                        .strip()
                    )
                    # When a video icon is placed between the link and speaker
                    if tmp_speaker in ("Watch Live", "Video"):
                        tmp_speaker = (
                            speech_link.parent.next_sibling.next_sibling.next_sibling.next_element.get_text()
                            .replace("\n", "")
                            .strip()
                        )
                self.speakers.append(tmp_speaker)
            if self.verbose:
                print("YEAR: {} - {} speeches found.".format(year, len(speech_links)))

    def _add_article(self, link, index=None):
        """
        Override a private function that adds a related article for 1 link into the instance variable
        The index is the index in the article to add to.
        Due to concurrent prcessing, we need to make sure the articles are stored in the right order
        """
        if self.verbose:
            sys.stdout.write(".")
            sys.stdout.flush()

        res = requests.get(self.base_url + link)
        html = res.text
        # p tag is not properly closed in many cases
        html = html.replace("<P", "<p").replace("</P>", "</p>")
        html = html.replace("<p", "</p><p").replace("</p><p", "<p", 1)
        # remove all after appendix or references
        x = re.search(
            r"(<b>references|<b>appendix|<strong>references|<strong>appendix)",
            html.lower(),
        )
        if x:
            html = html[: x.start()]
            html += "</body></html>"
        # Parse html text by BeautifulSoup
        article = BeautifulSoup(html, "html.parser")
        # Remove footnote
        for fn in article.find_all("a", {"name": re.compile("fn\d")}):
            if fn.parent:
                fn.parent.decompose()
            else:
                fn.decompose()
        # Get all p tag
        paragraphs = article.findAll("p")
        para_text = []
        for paragraph in paragraphs:
            text = re.sub(r"\s+", " ", paragraph.get_text().strip())
            if len(text) > 0 and text not in self.skip_text_list:
                para_text.append(text)
        self.articles[index] = self.segment_separator.join(para_text)



def load_text(path, num_heads=0, num_samples=-1):
    lines = []
    with open(path, encoding="utf-8") as f:
        if num_heads > 0:
            for _ in range(num_heads):
                next(f)
        for i, line in enumerate(f):
            if (num_samples > 0) and (i >= num_samples):
                break
            lines.append(line.rstrip("\n"))
    return lines


def load_parallel_text(source_path, target_path, num_heads=0, num_samples=-1):
    sources = load_text(source_path, num_heads, num_samples)
    targets = load_text(target_path, num_heads, num_samples)
    if len(sources) != len(targets):
        raise ValueError("Parallel corpus must have same length two files")
    return sources, targets


def load_wikitext(path, num_lines=-1):
    """
    Wikitext format

         = Head1 =

        text ...
        text ...

         = = 2ead = =

        text ...
        text ...
    """
    if num_lines <= 0:
        with open(path, encoding="utf-8") as f:
            texts = f.read().split("\n =")
    else:
        lines = []
        with open(path, encoding="utf-8") as f:
            for i, line in enumerate(f):
                if i >= num_lines:
                    break
                lines.append(line)
        texts = "".join(lines).split("\n =")
    # fix missing prefix
    texts = [texts[0]] + [f" ={text}" for text in texts[1:]]
    return texts



from typing import Any

from torch.utils.tensorboard import SummaryWriter

from snorkel.types import Config

from .log_writer import LogWriter


class TensorBoardWriter(LogWriter):
    """A class for logging to Tensorboard during training process.

    See ``LogWriter`` for more attributes.

    Parameters
    ----------
    kwargs
        Passed to ``LogWriter`` initializer

    Attributes
    ----------
    writer
        ``SummaryWriter`` for logging and visualization
    """

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.writer = SummaryWriter(self.log_dir)

    def add_scalar(self, name: str, value: float, step: float) -> None:
        """Log a scalar variable to TensorBoard.

        Parameters
        ----------
        name
            Name of the scalar collection
        value
            Value of scalar
        step
            Step axis value
        """
        self.writer.add_scalar(name, value, step)

    def write_config(
        self, config: Config, config_filename: str = "config.json"
    ) -> None:
        """Dump the config to file and add it to TensorBoard.

        Parameters
        ----------
        config
            JSON-compatible config to write to TensorBoard
        config_filename
            File to write config to
        """
        super().write_config(config, config_filename)
        self.writer.add_text(tag="config", text_string=str(config))

    def cleanup(self) -> None:
        """Close the ``SummaryWriter``."""
        self.writer.close()



import os
import zstandard
import orjson as json
import time
import tarfile
import codecs
from functools import reduce
import jsonlines
import io
from zipfile import ZipFile
import gzip
from math import ceil
import mmap
import multiprocessing as mp
from pathlib import Path

VALID_EXTENSIONS = [
    "openwebtext.tar.xz",
    "_data.xz",
    ".dat.zst",
    ".jsonl",
    ".jsonl.zst",
    ".jsonl.zst.tar",
    ".json.zst",
    ".txt",
    ".zip",
    ".tar.gz",
    ".json.gz",
    ".gz",
]


def has_valid_extension(file):
    return any([file.endswith(ext) for ext in VALID_EXTENSIONS])


def _listdir_or_file(x):
    if isinstance(x, list):
        return reduce(lambda x, y: x + y, map(listdir_or_file, sorted(x)))
    if os.path.isfile(x):
        return [x]
    elif os.path.isdir(x):
        return [str(Path(x) / fn) for fn in sorted(os.listdir(x))]
    else:
        raise FileNotFoundError(f"{x} not found")


def listdir_or_file(x):
    return list(filter(has_valid_extension, _listdir_or_file(x)))


def tarfile_reader(file, streaming=False):
    # we need our own tarfile parser because `tarfile` doesn't work well for
    # big tarfiles; it seems to be reading the entire file to get a list of
    # where all the files are - but we don't need that because we just need
    # to see each file once. surprisingly, `tarfile` doesn't expose any
    # facilities for this. the only options are 1. load the entire tarfile
    # and then query by filename or 2. extract to disk - and neither of
    # these is what we want.

    offset = 0
    paxfilesize = None
    while True:
        hdr = file.read(512)
        offset += 512

        # https://www.gnu.org/software/tar/manual/html_node/Standard.html
        # end at 135 not 136 because of \0 terminator
        if hdr[124:135] == b"\0" * 11:
            # end of record
            break

        fname = hdr[:100].split(b"\0")[0]

        # if the file is too big to fit in the size field, tarfiles will actually
        # include a PaxHeader with the size in it, applicable to the immediate next file.
        if paxfilesize is not None:
            size = paxfilesize
            paxfilesize = None
        else:
            size = int(hdr[124:135], 8)

        padded_size = ceil(size / 512) * 512

        # for handling PaxHeader files (which contain extra metadata about file size) and directories
        # https://pubs.opengroup.org/onlinepubs/9699919799/utilities/pax.html#tag_20_92_13_03
        type = chr(hdr[156])

        if type == "x":
            meta = file.read(padded_size)[:size]

            def kv(x):
                return x.decode("utf-8").split(" ")[1].split("=")

            paxfileattrs = {kv(x)[0]: kv(x)[1] for x in meta.split(b"\n") if x}
            paxfilesize = int(paxfileattrs["size"])

            offset += padded_size
            continue
        elif type != "0" and type != "\0":
            if streaming:
                file.seek(padded_size, os.SEEK_CUR)
            else:
                file.read(padded_size)
            offset += padded_size
            continue

        if streaming:
            # skip directory entries
            if size != 0:
                mmo = mmap.mmap(
                    file.fileno(), length=offset + size, access=mmap.ACCESS_READ
                )
                mmo.seek(offset)
                yield mmo

            file.seek(padded_size, os.SEEK_CUR)
        else:
            yield file.read(padded_size)[:size]
        offset += padded_size


def handle_jsonl(jsonl_reader, get_meta, autojoin_paragraphs, para_joiner, key="text"):
    for ob in jsonl_reader:
        # naive jsonl where each object is just the string itself, with no meta. For legacy compatibility.
        if isinstance(ob, str):
            assert not get_meta
            yield ob
            continue

        text = ob[key]

        if autojoin_paragraphs and isinstance(text, list):
            text = para_joiner.join(text)

        if get_meta:
            yield text, (ob["meta"] if "meta" in ob else {})
        else:
            yield text


class Reader:
    def __init__(self, in_path):
        self.in_path = in_path

    def stream_data(self, get_meta=False, threaded=False):
        if not threaded:
            yield from self._stream_data(get_meta)
            return

        q = mp.Queue(1000)
        p = mp.Process(target=self._stream_data_threaded, args=(q, get_meta))
        p.start()
        while p.is_alive():
            res = q.get()
            if res is None:
                break
            yield res

    def _stream_data_threaded(self, q, get_meta=False):
        for data in self._stream_data(get_meta):
            q.put(data)
        q.put(None)

    def _stream_data(self, get_meta=False, jsonl_key="text"):
        self.f_name = ""
        files = listdir_or_file(self.in_path)
        if not files:
            raise FileNotFoundError(f"No valid file(s) found in {self.in_path}")
        for f in files:
            self.f_name = f
            if f == "openwebtext.tar.xz":
                assert not get_meta

                yield from self.read_owt(f)
            elif "urlsf_subset" in f and f.endswith("_data.xz"):
                assert not get_meta

                yield from self.read_owt_subset(f)
            elif f.endswith(".dat.zst"):
                assert not get_meta

                yield from self.read_dat(f)
            elif f.endswith(".jsonl"):
                yield from self.read_jsonl(f, get_meta, key=jsonl_key)
            elif f.endswith(".jsonl.zst"):
                yield from self.read_jsonl_zst(f, get_meta, key=jsonl_key)
            elif f.endswith(".jsonl.zst.tar"):
                yield from self.read_jsonl_tar(f, get_meta, key=jsonl_key)
            elif f.endswith(".json.zst"):
                assert not get_meta

                yield from self.read_json(f)
            elif f.endswith(".txt"):
                assert not get_meta

                yield from self.read_txt(f)
            elif f.endswith(".zip"):
                assert not get_meta

                yield from self.read_zip(f)
            elif f.endswith(".tar.gz"):
                assert not get_meta

                yield from self.read_tgz(f)
            elif f.endswith(".json.gz"):
                assert not get_meta

                yield from self.read_jsongz(f)
            elif f.endswith(".gz"):
                assert not get_meta

                yield from self.read_gz(f)
            else:
                # shouldn't be reached
                print(f"Skipping {f} as streaming for that filetype is not implemented")

    def read_txt(self, file):
        with open(file, "r") as fh:
            yield fh.read()

    def read_zip(self, file):
        archive = ZipFile(file, "r")
        for f in archive.namelist():
            yield archive.read(f).decode("UTF-8")

    def read_tgz(self, file):
        gz = gzip.open(file)
        yield from (x.decode("utf-8") for x in tarfile_reader(gz, streaming=False))

    def read_gz(self, file):
        with gzip.open(file, "rb") as f:
            for line in f:
                yield line.decode("utf-8")

    def read_jsongz(self, file):
        for line in self.read_gz(file):
            yield json.loads(line)

    def read_json(self, file):
        with open(file, "rb") as fh:
            cctx = zstandard.ZstdDecompressor()
            reader = cctx.stream_reader(fh)
            ob = json.load(reader)
            yield from ob

    def read_dat(self, file):
        with open(file, "rb") as fh:
            cctx = zstandard.ZstdDecompressor()
            reader = cctx.stream_reader(fh)
            while True:
                ln = reader.read(16).decode("UTF-8")
                if not ln:
                    break

                ln = int(ln)

                yield reader.read(ln).decode("UTF-8")

    def read_jsonl(
        self,
        file,
        get_meta=False,
        autojoin_paragraphs=True,
        para_joiner="\n\n",
        key="text",
    ):
        with jsonlines.open(file) as rdr:
            yield from handle_jsonl(
                rdr, get_meta, autojoin_paragraphs, para_joiner, key
            )

    def read_jsonl_zst(
        self,
        file,
        get_meta=False,
        autojoin_paragraphs=True,
        para_joiner="\n\n",
        key="text",
    ):
        with open(file, "rb") as fh:
            cctx = zstandard.ZstdDecompressor()
            reader = io.BufferedReader(cctx.stream_reader(fh))
            rdr = jsonlines.Reader(reader)
            yield from handle_jsonl(
                rdr, get_meta, autojoin_paragraphs, para_joiner, key
            )

    def read_jsonl_tar(
        self,
        file,
        get_meta=False,
        autojoin_paragraphs=True,
        para_joiner="\n\n",
        key="text",
    ):
        with open(file, "rb") as fh:
            for f in tarfile_reader(fh, streaming=True):
                cctx = zstandard.ZstdDecompressor()
                reader = io.BufferedReader(cctx.stream_reader(f))
                rdr = jsonlines.Reader(reader)
                yield from handle_jsonl(
                    rdr, get_meta, autojoin_paragraphs, para_joiner, key
                )
                f.close()

    def read_owt(self, file):
        tar = tarfile.open(file, encoding="utf-8")
        utf8reader = codecs.getreader("utf-8")

        for name in tar.getmembers():
            fp = tar.extractfile(name)
            inner_tar = tarfile.open(fileobj=fp, encoding="utf-8")
            for inner_name in inner_tar.getmembers():
                inner_fp = utf8reader(inner_tar.extractfile(inner_name))
                contents = inner_fp.read()
                yield contents

    def read_owt_subset(self, file):
        utf8reader = codecs.getreader("utf-8")
        tar = tarfile.open(file, encoding="utf-8")
        for name in tar.getmembers():
            fp = utf8reader(tar.extractfile(name))
            contents = fp.read()
            yield contents


class Archive:
    def __init__(self, out_dir, compression_level=3):
        self.out_dir = out_dir
        os.makedirs(out_dir, exist_ok=True)
        self.i = 0

        self.fh = open(self.out_dir + "/current_chunk_incomplete", "wb")
        self.cctx = zstandard.ZstdCompressor(level=compression_level, threads=8)
        self.compressor = self.cctx.stream_writer(self.fh)

    def add_data(self, data, meta={}):
        self.compressor.write(
            json.dumps({"text": data, "meta": meta}).encode("UTF-8") + b"\n"
        )

    def commit(self, archive_name="default"):
        fname = (
            self.out_dir
            + "/data_"
            + str(self.i)
            + "_time"
            + str(int(time.time()))
            + "_"
            + archive_name
            + ".jsonl.zst"
        )
        self.compressor.flush(zstandard.FLUSH_FRAME)

        self.fh.flush()
        self.fh.close()
        os.rename(self.out_dir + "/current_chunk_incomplete", fname)
        self.fh = open(self.out_dir + "/current_chunk_incomplete", "wb")
        self.compressor = self.cctx.stream_writer(self.fh)

        self.i += 1


class DatArchive:
    def __init__(self, out_dir):
        self.out_dir = out_dir
        os.makedirs(out_dir, exist_ok=True)
        self.data = []
        self.i = 0
        if os.path.exists(out_dir) and len(os.listdir(out_dir)) > 0:
            self.i = (
                max(
                    map(
                        lambda x: int(x.split("_")[1].split(".")[0]),
                        os.listdir(out_dir),
                    )
                )
                + 1
            )

    def add_data(self, data):
        self.data.append(data)

    def commit(self, archive_name=None):
        # TODO: streaming
        cctx = zstandard.ZstdCompressor(level=3)

        if archive_name is None:
            archive_name = str(int(time.time()))

        res = b"".join(
            map(
                lambda x: ("%016d" % len(x)).encode("UTF-8") + x,
                map(lambda x: x.encode("UTF-8"), self.data),
            )
        )
        cdata = cctx.compress(res)

        with open(
            self.out_dir + "/data_" + str(self.i) + "_" + archive_name + ".dat.zst",
            "wb",
        ) as fh:
            fh.write(cdata)

        self.i += 1
        self.data = []


class JSONArchive:
    def __init__(self, out_dir):
        self.out_dir = out_dir
        os.makedirs(out_dir, exist_ok=True)
        self.data = []
        self.i = 0
        if os.path.exists(out_dir) and len(os.listdir(out_dir)) > 0:
            self.i = (
                max(
                    map(
                        lambda x: int(x.split("_")[1].split(".")[0]),
                        os.listdir(out_dir),
                    )
                )
                + 1
            )

    def add_data(self, data):
        self.data.append(data)

    def commit(self):
        cctx = zstandard.ZstdCompressor(level=3)

        cdata = cctx.compress(json.dumps(self.data).encode("UTF-8"))
        with open(
            self.out_dir
            + "/data_"
            + str(self.i)
            + "_"
            + str(int(time.time()))
            + ".json.zst",
            "wb",
        ) as fh:
            fh.write(cdata)

        self.i += 1
        self.data = []



import abc
import collections
import orjson as json
import os
import time
import random

from datasets import load_dataset
from pytablewriter import MarkdownTableWriter
from tqdm.auto import tqdm

from ekorpkit.utils.func import humanbytes, parse_size, utf8len

from ekorpkit import eKonf
from ekorpkit.utils.func import change_directory
from hydra.utils import instantiate

from tqdm.auto import tqdm
import zstandard


class Archive:
    def __init__(self, output_dir, compression_level=3):
        self.output_dir = output_dir
        os.makedirs(output_dir, exist_ok=True)
        self.i = 0

        self.fh = open(self.output_dir + "/current_chunk_incomplete", "wb")
        self.cctx = zstandard.ZstdCompressor(level=compression_level, threads=8)
        self.compressor = self.cctx.stream_writer(self.fh)

    def add_data(self, data, meta={}):
        self.compressor.write(
            json.dumps({"text": data, "meta": meta}).encode("UTF-8") + b"\n"
        )

    def commit(self, archive_name="default"):
        fname = (
            self.output_dir
            + "/data_"
            + str(self.i)
            + "_time"
            + str(int(time.time()))
            + "_"
            + archive_name
            + ".jsonl.zst"
        )
        self.compressor.flush(zstandard.FLUSH_FRAME)

        self.fh.flush()
        self.fh.close()
        os.rename(self.output_dir + "/current_chunk_incomplete", fname)
        self.fh = open(self.output_dir + "/current_chunk_incomplete", "wb")
        self.compressor = self.cctx.stream_writer(self.fh)

        self.i += 1


def dummy_meta(xs):
    return ((x, {}) for x in xs)


class Dataset(abc.ABC):
    @abc.abstractmethod
    def name(self):
        """Human-readable name of the dataset"""
        pass

    @abc.abstractmethod
    def documents(self):
        """A generator producing all documents in the dataset."""
        pass

    @abc.abstractmethod
    def clean(self):
        """Remove any dataset files."""
        pass

    def size(self):
        """Return an estimate of the dataset size. Implementations may use a faster, less accurate estimate."""

        size = sum(map(utf8len, tqdm(self.documents())))
        print("size", self.name(), size)
        return size

    def num_docs(self):
        """Return an estimate of the number of documents in the dataset. Implementations may use a faster, less accurate estimate."""

        size = len(list(map(lambda x: None, tqdm(self.documents()))))
        print("docs", self.name(), size)
        return size

    def shuffled(self):
        """Datasets where the source is already shuffled should override this to return True so that it isn't shuffled again."""
        return False


class HFDS(Dataset):
    def __init__(self, **args):
        self.name = args["name"]
        self.path = args["path"]
        self.subset = args["subset"]
        self.split = args[eKonf.Keys.SPLIT]
        self.download_mode = args["download_mode"]
        self.cache_dir = args["cache_dir"]
        self.streaming = args["streaming"]

    def name(self):
        return self.name

    def _load(self):
        self.dataset = load_dataset(
            self.name, self.subset, download_mode=self.download_mode
        )

    def documents(self):
        self._load()

        for doc in self.dataset:
            yield doc

    def clean(self):
        if not self.streaming:
            self.dataset.cleanup_cache_files()


def take(n, iter):
    ret = []
    for i in range(n):
        try:
            ret.append(next(iter))
        except StopIteration:
            break
    return ret


def mk_table(datasets, train_chars, latex=None, print_latex=False):
    values = []

    total_weight = sum([x[1] * x[0].size() for x in datasets])

    for dataset, weight in datasets:
        size = dataset.size()
        relative_weight = size * weight / total_weight
        values.append(
            [
                dataset.name(),
                size,
                "{:.2%}".format(relative_weight),
                "{:.4f}".format(train_chars / size * relative_weight),
                size * weight,
                humanbytes(size / dataset.num_docs(), "KiB"),
            ]
        )

    values.sort(key=lambda x: -x[4])
    values.append(
        [
            "**Total**",
            "",
            "",
            "",
            sum([x[4] for x in values]),
            humanbytes(
                sum([x[1] for x in values]) / sum(x[0].num_docs() for x in datasets),
                "KiB",
            ),
        ]
    )
    values = [
        [
            x[0],
            humanbytes(x[1], "GiB") if x[1] else "",
            x[2],
            x[3],
            humanbytes(x[4], "GiB"),
            x[5],
        ]
        for x in values
    ]

    writer = MarkdownTableWriter()
    writer.table_name = "The Pile™"
    writer.headers = [
        "Component",
        "Raw Size",
        "Weight",
        "Epochs",
        "Effective Size",
        "Mean Document Size",
    ]
    writer.value_matrix = values

    if print_latex:
        rows = []
        for row in values[:-1]:
            rows.append(
                "        "
                + " & ".join(map(lambda x: str(x).replace("%", r"\%"), row))
                + r" \\"
            )
        totalrow = (
            " & ".join(
                map(
                    lambda x: r"\textbf{%s}" % str(x).replace("%", r"\%") if x else "",
                    values[-1][1:],
                )
            )
            + r" \\"
        )
        latex = latex.format(rows="\n".join(rows), totalrow=totalrow)
        print(latex)
    return writer.dumps()


def dataset_tqdm(dset):
    if isinstance(dset, LMDS):
        return dset.documents()
    pbar = tqdm(total=dset.size(), unit="B", unit_scale=True, unit_divisor=1024)
    for doc in dset.documents():
        pbar.update(utf8len(doc))
        yield doc


class Profiler:
    def __init__(self, profile):
        self.i = 0
        self.profile = profile
        self.time_per_dataset = collections.defaultdict(lambda: [0, 0])

    def measured_next(self, name, iter):
        if not self.profile:
            # no-op
            return next(iter)
        else:
            self.i += 1
            start = time.time()
            doc = next(iter)
            elapsed = time.time() - start

            self.time_per_dataset[name][0] += elapsed
            self.time_per_dataset[name][1] += 1

            if self.i % 100000 == 0:
                times = [
                    (dsname, total, ct)
                    for dsname, (total, ct) in self.time_per_dataset.items()
                ]
                times.sort(key=lambda x: x[1])
                for name, total, ct in times:
                    print(
                        name.ljust(22),
                        "{:.8f}".format(total / ct),
                        str(ct).rjust(8),
                        "{:.4f}".format(total),
                    )

            return doc


class LMDS(Dataset):
    def __init__(self, datasets, dataset_bytes, profile=False):
        self.datasets = datasets
        self.dataset_bytes = dataset_bytes
        self.profile = profile
        self.rnd = random.Random(42)

    def name(self):
        return "Custom Pile"

    def documents(self):
        datasets = []
        weights = []

        # calculate relative_weight for each
        total_weight = sum([x[1] * x[0].num_docs() for x in self.datasets])
        for dataset, weight in self.datasets:
            size = dataset.size()
            relative_weight = weight * dataset.num_docs() / total_weight
            datasets.append((dataset.name(), dataset))
            weights.append(relative_weight)

        # yield from dataset until right number of bytes
        total_bytes = 0
        pbar = tqdm(
            total=self.dataset_bytes, unit="B", unit_scale=True, unit_divisor=1024
        )

        profiler = Profiler(profile=self.profile)
        while True:
            chunk = self.rnd.choices(population=datasets, weights=weights, k=1000)
            for name, dset in chunk:
                doc, meta = profiler.measured_next(name, dset)

                size = utf8len(doc)
                total_bytes += size
                pbar.update(size)

                meta["pile_set_name"] = name

                yield doc, meta

                if total_bytes > self.dataset_bytes:
                    return

    def clean(self):
        for dataset, _ in self.datasets:
            dataset.clean()

    def size(self):
        return self.dataset_bytes


class LimitedDataset(Dataset):
    def __init__(self, dataset, limit_size):
        self.dataset = dataset
        self.limit_size = limit_size
        self.rnd = random.Random(42)

    def name(self):
        return self.dataset.name() + " (truncated)"

    def documents(self):
        numer = self.limit_size
        denom = self.dataset.size()
        for doc, meta in dataset_tqdm(self.dataset):
            docsize = utf8len(doc)
            if self.rnd.random() < numer / denom:
                yield doc, meta
                numer -= docsize
            denom -= docsize

            if numer <= 0 or denom <= 0:
                break

    def clean(self):
        self.dataset.clean()

    def size(self):
        return self.limit_size


def sample_from_sets(datasets, n_docs):
    random.seed(42)
    for dset, _ in datasets:
        print(dset.name())
        fname = "dataset_samples/{}.json".format(dset.name().replace(" ", "_"))
        if os.path.exists(fname):
            continue

        n = dset.num_docs()

        indices = set(random.sample(range(n), n_docs))
        pbar = tqdm(total=n_docs)

        docs = []
        for i, (doc, meta) in enumerate(dset.documents()):
            if i > max(indices):
                break
            if i in indices:
                docs.append((doc, meta))
                pbar.update(1)

        try:
            os.mkdir("dataset_samples")
        except:
            pass

        with open(fname, "w") as fh:
            json.dump(docs, fh)

        pbar.close()


def build_pile(**args):
    args = eKonf.to_config(args)

    if args.output_dir:
        os.makedirs(args.output_dir, exist_ok=True)

    with change_directory(args.output_dir):
        main(args)


def main(args):

    if args.datasets:
        datasets = []
        for dataset in args.datasets:
            dataset = instantiate(dataset)
            datasets.append((dataset, 1.0))

    random.seed(42)

    # if args.using != 'pile_reprod_no_cc':
    #     # add CC
    #     datasets.append((CommonCrawlDataset(), 1.))

    if args.read_amount is None:
        args.read_amount = sum([ds.size() * epochs for ds, epochs in datasets])
    else:
        args.read_amount = parse_size(args.read_amount)

    print(mk_table(datasets, args.read_amount))

    lmd = LMDS(datasets, args.read_amount, profile=args.profile)

    if args.force_download:
        for dset, _ in datasets:
            dset._download()

    if args.limit:
        size_limit = parse_size(args.limit)
        lmd = LimitedDataset(lmd, size_limit)

    if args.make_lmd:
        assert not (args.interleave_output and args.chunk)  # can't chunk and interleave

        if args.interleave_output:
            ars = [
                Archive("pile_pass1/chunk{}".format(i))
                for i in range(args.interleave_output)
            ]
        else:
            ar = Archive("pile_output")

        if args.chunk:
            chunk_size = parse_size(args.chunk)

        cursize = 0
        for doc, meta in lmd.documents():
            if args.interleave_output:
                ar = random.choice(ars)

            ar.add_data(doc, meta)

            cursize += len(doc)
            if args.chunk and cursize > chunk_size:
                # interleave will not be on
                cursize = 0
                ar.commit(archive_name=args.using)

        if args.interleave_output:
            for ar in ars:
                ar.commit(archive_name=args.using)
        else:
            ar.commit(archive_name=args.using)

    if args.make_dataset_samples:
        sample_from_sets(datasets, args.make_dataset_samples)



import logging
import pandas as pd
from tqdm.auto import tqdm
from ekorpkit import eKonf
from .ngram import Ngrams
from .score import get_process_memory, prune_vocab


log = logging.getLogger(__name__)


class NgramTrainer(Ngrams):
    def __init__(
        self,
        **args,
    ):
        args = eKonf.to_config(args)
        self._candidates = args.candidates
        if self._candidates.min_count <= 0:
            self._candidates.min_count = 10
        if self._candidates.max_candidates <= 0:
            self._candidates.max_candidates = 40000000

        self.score_function = eKonf.partial(args.score_function)
        self._ngrams = {}
        self._total_words = 0

        super().__init__(**args)

    def save_candidates(self):
        """Save the candidates to a file"""
        if len(self.candidates) > 0:
            df = pd.DataFrame(self.candidates.values())
            eKonf.save_data(df, self.score_path, verbose=self.verbose)
        else:
            log.info("no candidates to save")

    def train(self):
        """Train the model"""
        if not self._sentences:
            self.load_data()
        self.learn_ngrams()
        self.score_ngrams()

    def score_ngrams(self):
        if self.score_function is not None and callable(self.score_function):
            self.candidates = self.score_function(
                self._ngrams, total_words=self._total_words
            )

    def learn_ngrams(self):

        self._ngrams = {}
        sentence_no, total_words = -1, 0
        for sentence_no, sentence in tqdm(enumerate(self._sentences)):
            words = self.tokenize(sentence)
            total_words += len(words)

            ngrams = self.prepare_ngram_tuples(
                words,
                max_n=self._ngram.max_n,
                max_window=self._ngram.max_window,
                max_skip=self._ngram.max_skip,
            )
            for ngram in ngrams:
                self._ngrams[ngram] = self._ngrams.get(ngram, 0) + 1

            if (
                (sentence_no > 0)
                and (self._candidates.num_sents_per_pruning > 0)
                and (sentence_no % self._candidates.num_sents_per_pruning == 0)
            ):
                prune_vocab(self._ngrams, self._candidates.prune_min_ngram_count)
            if len(self._ngrams) > self._candidates.max_candidates:
                prune_vocab(self._ngrams, self._candidates.prune_min_ngram_count)
                self._candidates.prune_min_ngram_count += 1

        if self.verbose:
            print("learning ngrams was done. memory= %.3f Gb" % get_process_memory())

        self._ngrams = {
            ngram: freq
            for ngram, freq in self._ngrams.items()
            if freq >= self._candidates.min_count
        }
        self._total_words = total_words



import pandas as pd
from pathlib import Path
from bs4 import BeautifulSoup
from tqdm.auto import tqdm
from joblib import Parallel, delayed
from ekorpkit.utils.batch.batcher import tqdm_joblib


class EarningsCall:
    def __init__(
        self,
        name,
        input_file,
        output_dir,
        output_file,
        filetype,
        force_download,
        num_workers,
        **kwargs,
    ):
        import tarfile

        output_dir = Path(output_dir)
        if not output_dir.is_dir():
            output_dir.mkdir(exist_ok=True)
        output_path = output_dir / output_file

        if output_path.exists() and not force_download:
            print(f"Corpus [{name}] is already built to [{output_path}]")
            return

        documents = []
        with tarfile.open(input_file) as tar:
            files = [file for file in tar.getmembers() if filetype in file.name]
            print(f"Number of files in [{input_file}]: {len(files)}")
            processes = num_workers if num_workers else 1
            desciption = f"Extracting {filetype} files"
            with tqdm_joblib(tqdm(desc=desciption, total=len(files))) as pbar:
                results = Parallel(n_jobs=processes)(
                    delayed(html_to_document)(tar.extractfile(file).read(), file.name)
                    for file in files
                )
                for result in results:
                    documents.append(result)

            # pool = mp.Pool(processes=processes)
            # results = []
            # for file in tqdm(files):
            # 	data = tar.extractfile(file).read()
            # 	results.append(pool.apply_async(html_to_document, args=(data, file.name)))
            # pool.close()
            # pool.join()
            # for result in results:
            # 	doc = result.get()
            # 	documents.append(doc)

        reports_df = pd.DataFrame(documents)
        reports_df.to_csv(output_path, header=True)
        print(reports_df.tail())
        print(f"Corpus [{name}] is built to [{output_path}] from [{input_file}]")


def html_to_document(data, filename):
    soup = BeautifulSoup(data, "html.parser")
    ps = soup.find_all("p")
    text = ""
    title = ""
    for i, p in enumerate(ps):
        if i == 0:
            title = p.text.strip()
        else:
            text += p.text + "\n"
    doc = {
        "filename": filename,
        "title": title,
        "nos": len(text.split("\n")),
        "text": text,
    }
    return doc
